import glob
import os
import shlex
import shutil
import socket
import subprocess
import textwrap

import click

from gcs_cli.helpers import is_logged_in
from gcs_cli.parsing import globus_command


def _run_diagnostic_step(cmd, force_try_all):
    stagename = (
        str(cmd)
        if not callable(cmd)
        else getattr(cmd, "stagename", "func:{}()".format(cmd.__name__))
    )
    click.secho("== diagnostic: {} ==".format(stagename), fg="yellow", bold=True)

    try:
        if callable(cmd):
            result = cmd()
            if isinstance(result, str):
                click.echo(result)
            return

        if isinstance(cmd, str):
            cmd = shlex.split(cmd)
        arg0 = cmd[0]
        if force_try_all or shutil.which(arg0):
            try:
                subprocess.run(cmd, timeout=30)
            except subprocess.TimeoutExpired:
                click.secho("command timed out", fg="red", bold=True)
        else:
            click.secho("{} was not found in the PATH".format(arg0), fg="red")
    # at the very least, subprocess execution can raise an OSError
    # when attempting to execute a non-executable file or when the command name
    # can't be resolved
    # any errors here should be caught to avoid crashing the whole diagnostic run
    except Exception as e:
        click.secho("command failed: {}".format(e), fg="red", bold=True)


def _is_debian():
    """return true if the platform is a debian-like system
    false if it is not or if detection fails in any way"""
    try:
        with open("/etc/os-release", "r") as f:
            for line in f:
                if line.startswith("ID_LIKE"):
                    return line.strip() == "ID_LIKE=debian"
    except Exception:
        pass
    return False


def cat(path, wildcard=False):
    def show_step():
        if not wildcard:
            files = [path]
        else:
            files = glob.glob(path)

        for filename in files:
            click.echo("cat " + filename)
            if os.path.exists(filename):
                with open(filename) as f:
                    click.echo(textwrap.indent(f.read(), "  "))
            else:
                click.echo("no file named {}".format(filename))

    show_step.stagename = "func:cat({})".format(path)

    return show_step


def systemctl_status(service):
    return "systemctl status {} --no-pager --lines 10".format(service)


def get_fqdn():
    return socket.getfqdn()


def user_is_logged_in():
    return str(is_logged_in())


def running_as_root():
    return os.geteuid() == 0


def get_service_names(force_try_all):
    services = ["gcs_manager", "gcs_manager_assistant", "globus-gridftp-server"]
    # add apache2 or httpd depending on platform in use
    if force_try_all:
        return ["apache2", "httpd"] + services
    if _is_debian():
        return ["apache2"] + services
    return ["httpd"] + services


@globus_command("self-diagnostic", disable_options=["map_http_status", "format"])
@click.option(
    "--force-try-all",
    is_flag=True,
    help="Don't skip any steps that are expected to fail",
)
def self_diagnostic_command(force_try_all):
    """Run several diagnostic utilities which can be used to provide feedback to the
    Globus Team.

    This tool may produce a large amount of output.
    """
    any_user_diagnostics = [
        "uname -a",
        "sestatus",
        "ifconfig",
        get_fqdn,
        cat("/etc/os-release"),
        cat("/var/lib/globus-connect-server/info.json"),
        user_is_logged_in,
        "globus-gridftp-server --version",
    ] + [systemctl_status(x) for x in get_service_names(force_try_all)]
    root_user_diagnostics = [
        cat("/etc/gridftp.conf"),
        cat("/etc/gridftp.d/*", wildcard=True),
    ]

    for cmd in any_user_diagnostics:
        _run_diagnostic_step(cmd, force_try_all)

    is_root = running_as_root()
    click.secho("= check running_as_root:{} =".format(is_root), fg="yellow", bold=True)
    if is_root or force_try_all:
        for cmd in root_user_diagnostics:
            _run_diagnostic_step(cmd, force_try_all)
