import json

from gcs_cli.helpers import filter_fields, requires_login, sorted_json_field
from gcs_cli.parsing import (
    globus_command,
    include_private_policies_opt,
    storage_gateway_id_arg,
)
from gcs_cli.parsing.stateful_options import outformat_is_text
from gcs_cli.prettyprint import FORMAT_TEXT_RAW, FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.services.gcs import (
    BLACKPEARL_CONNECTOR_ID,
    BOX_CONNECTOR_ID,
    CEPH_CONNECTOR_ID,
    GOOGLE_CLOUD_STORAGE_CONNECTOR_ID,
    GOOGLE_DRIVE_CONNECTOR_ID,
    POSIX_CONNECTOR_ID,
    S3_CONNECTOR_ID,
    connector_id_to_display_name,
    get_client,
)

COMMON_FIELDS = [
    ("Display Name", "display_name"),
    ("ID", "id"),
    ("Connector", lambda x: connector_id_to_display_name(x["connector_id"])),
    ("High Assurance", "high_assurance"),
    ("Authentication Timeout", "authentication_timeout_mins"),
    ("Allowed Domains", "allowed_domains"),
]
PRIVATE_COMMON_FIELDS = [
    ("Path Restrictions", sorted_json_field("restrict_paths")),
    ("Allowed Users", "users_allow"),
    ("Denied Users", "users_deny"),
]

POSIX_FIELDS = []
PRIVATE_POSIX_FIELDS = [
    ("Allowed POSIX Groups", "groups_allow"),
    ("Denied POSIX Groups", "groups_deny"),
]

BLACKPEARL_FIELDS = [
    ("S3 Endpoint URL", "s3_endpoint"),
    ("BlackPearl Access File", "bp_access_id_file"),
]
PRIVATE_BLACKPEARL_FIELDS = []

BOX_FIELDS = [("Enterprise ID", "enterpriseID")]
PRIVATE_BOX_FIELDS = [("Box App Settings", "boxAppSettings")]

CEPH_FIELDS = [("S3 Endpoint URL", "s3_endpoint"), ("S3 Buckets", "s3_buckets")]
PRIVATE_CEPH_FIELDS = [
    ("Ceph Admin Key ID", "ceph_admin_key_id"),
    ("Ceph Admin Secret Key", "ceph_admin_secret_key"),
]

GOOGLE_DRIVE_FIELDS = []
PRIVATE_GOOGLE_DRIVE_FIELDS = [
    ("Google Client ID", "client_id"),
    ("Google Client Secret", "secret"),
]

GOOGLE_CLOUD_STORAGE_FIELDS = [
    ("Google Cloud Storage Buckets", "buckets"),
    ("Google Cloud Storage Projects", "projects"),
]
PRIVATE_GOOGLE_CLOUD_STORAGE_FIELDS = [
    ("Google Client ID", "client_id"),
    ("Google Client Secret", "secret"),
]

S3_FIELDS = [
    ("S3 Endpoint URL", "s3_endpoint"),
    ("S3 Buckets", "s3_buckets"),
    ("S3 User Credential Required", "s3_user_credential_required"),
]
PRIVATE_S3_FIELDS = []


@globus_command("show", short_help="Show a storage gateway definition")
@include_private_policies_opt
@storage_gateway_id_arg
@requires_login
def storage_gateway_show(storage_gateway_id, include_private_policies):
    """
    Show a storage gateway definition by id on the endpoint the GCS cli is logged into.
    """
    client = get_client()

    # output connector information
    include = []
    if include_private_policies:
        include = ["private_policies"]
    res = client.get_storage_gateway(storage_gateway_id, include=include)

    # filter only to fields which are present
    fields = filter_fields(COMMON_FIELDS, res)
    if include_private_policies:
        fields += filter_fields(PRIVATE_COMMON_FIELDS, res)

    formatted_print(res, text_format=FORMAT_TEXT_RECORD, fields=fields)

    # output connector policy information
    if outformat_is_text():
        policy_fields = None

        connector_id = res["connector_id"]

        if connector_id == POSIX_CONNECTOR_ID:
            policy_fields = POSIX_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_POSIX_FIELDS
        elif connector_id == BLACKPEARL_CONNECTOR_ID:
            policy_fields = BLACKPEARL_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_BLACKPEARL_FIELDS
        elif connector_id == BOX_CONNECTOR_ID:
            policy_fields = BOX_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_BOX_FIELDS
        elif connector_id == CEPH_CONNECTOR_ID:
            policy_fields = CEPH_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_CEPH_FIELDS
        elif connector_id == GOOGLE_DRIVE_CONNECTOR_ID:
            policy_fields = GOOGLE_DRIVE_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_GOOGLE_DRIVE_FIELDS
        elif connector_id == GOOGLE_CLOUD_STORAGE_CONNECTOR_ID:
            policy_fields = GOOGLE_CLOUD_STORAGE_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_GOOGLE_CLOUD_STORAGE_FIELDS
        elif connector_id == S3_CONNECTOR_ID:
            policy_fields = S3_FIELDS
            if include_private_policies:
                policy_fields += PRIVATE_S3_FIELDS
        else:
            formatted_print(
                json.dumps(res["policies"], sort_keys=True),
                text_format=FORMAT_TEXT_RAW,
                text_preamble="\nRaw JSON policies from unknown Connector:",
            )

        policy_fields = filter_fields(policy_fields, res["policies"])

        if policy_fields:
            formatted_print(
                res["policies"],
                text_format=FORMAT_TEXT_RECORD,
                fields=policy_fields,
                text_preamble="\nConnector Specific Policies:",
            )
