import logging.config
import os

from globus_sdk_tokenstorage import SQLiteAdapter

import globus_sdk

__all__ = [
    # option name constants
    "GLOBUS_ENV",
    "GLOBUS_PROFILE",
    "internal_native_client",
    "internal_auth_client",
    "token_storage_adapter",
]

# configuration names stored and retrieved on the globus_sdk_tokenstorage.SQLiteAdapter
# using store_config and read_config
AUTH_CLIENT_DATA_CONFIG_KEY = "auth_client_data"
# TODO: right now, this uses a singular "gcs_client_data" because the CLI only
# talks to one GCS at a time
# this is baked into login and logout, but it doesn't really belong there in
# this way
# instead, we need to be storing and using multiple resource servers' config data (for
# now, just FQDN) -- so perhaps this will become a family of string and we'll
# store under
#   "gcs_client_data_{client_id}"
# as the key?
#
# the impact on logout, in particular, needs to be figured out
GCS_CLIENT_DATA_CONFIG_KEY = "gcs_client_data"

# get the environment from env var (not exported)
GLOBUS_ENV = os.environ.get("GLOBUS_SDK_ENVIRONMENT")

# get profile name (allows you to sign in with multiple users)
GLOBUS_PROFILE = os.environ.get("GLOBUS_PROFILE", "default")


# config for use whenever we need to do things using
# instance clients from the CLI Native App Template
# primarily accessed via `internal_auth_client()`
def _get_template_client_id():
    default_id = "95fdeba8-fac2-42bd-a357-e068d82ff78e"
    return {
        "sandbox": "33b6a241-bce4-4359-9c6d-09f88b3c9eef",
        "integration": "e0c31fd1-663b-44e1-840f-f4304bb9ee7a",
        "test": "0ebfd058-452f-40c3-babf-5a6b16a7b337",
        "staging": "3029c3cb-c8d9-4f2b-979c-c53330aa7327",
        "preview": "b2867dbb-0846-4579-8486-dc70763d700b",
    }.get(GLOBUS_ENV, default_id)


def _get_tokenstorage_db():
    storagedb_dir = os.path.expanduser("~/.globus/gcs_cli/")
    os.makedirs(storagedb_dir, exist_ok=True)
    return os.path.join(storagedb_dir, "tokenstore.db")


def _get_tokenstorage_namespace():
    if "/" in GLOBUS_PROFILE:
        raise ValueError("Not allowed to have '/' in GLOBUS_PROFILE!")
    if GLOBUS_ENV:
        return GLOBUS_ENV + "/" + GLOBUS_PROFILE
    else:
        return GLOBUS_PROFILE


def token_storage_adapter():
    if not hasattr(token_storage_adapter, "_instance"):
        # namespace is equal to the current environment
        token_storage_adapter._instance = SQLiteAdapter(
            _get_tokenstorage_db(), namespace=_get_tokenstorage_namespace()
        )
    return token_storage_adapter._instance


def load_client_data():
    adapter = token_storage_adapter()
    conf = adapter.read_config(AUTH_CLIENT_DATA_CONFIG_KEY)
    return conf or {}


def internal_native_client():
    template_id = load_client_data().get("template_id", _get_template_client_id())
    return globus_sdk.NativeAppAuthClient(template_id)


def internal_auth_client(force_new_client=False):
    """
    Looks up the values for this CLI's Instance Client in config

    If none exists or force_new_client is True,
    registers a new Instance Client with GLobus Auth

    Returns a ConfidentialAppAuthClient
    """
    client_data = load_client_data()
    client_id = client_data.get("client_id")
    client_secret = client_data.get("client_secret")
    template_id = client_data.get("template_id", _get_template_client_id())
    template_client = internal_native_client()
    existing = client_id and client_secret

    # if we are forcing a new client, delete any existing client
    if force_new_client and existing:
        existing_client = globus_sdk.ConfidentialAppAuthClient(client_id, client_secret)
        try:
            existing_client.delete("/v2/api/clients/{}".format(client_id))

        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except globus_sdk.exc.AuthAPIError:
            pass

    # if we require a new client to be made
    if force_new_client or not existing:
        # register a new instance client with auth
        body = {"client": {"template_id": template_id, "name": "Globus CLI"}}
        res = template_client.post("/v2/api/clients", json_body=body)

        # get values and write to config
        credential_data = res["included"]["client_credential"]
        client_id = client_data["client_id"] = credential_data["client"]
        client_secret = client_data["client_secret"] = credential_data["secret"]
        token_storage_adapter().store_config(AUTH_CLIENT_DATA_CONFIG_KEY, client_data)

    return globus_sdk.ConfidentialAppAuthClient(
        client_id, client_secret, app_name="Globus Connect Server CLI"
    )


def setup_logging(level="DEBUG"):
    conf = {
        "version": 1,
        "formatters": {
            "basic": {"format": "[%(levelname)s] %(name)s::%(funcName)s() %(message)s"}
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "basic",
            }
        },
        "loggers": {"globus_sdk": {"level": level, "handlers": ["console"]}},
    }

    logging.config.dictConfig(conf)
