import json

import click

import globus_sdk

from gcs_cli.parsing.stateful_options import outformat_is_text
from gcs_cli.sdk_client import GCSAPIError
from gcs_cli.services.auth import get_client


def handle_excinfo(excinfo):
    exception_type, exception, _ = excinfo
    if isinstance(
        exception, (click.ClickException, click.exceptions.Abort, click.exceptions.Exit)
    ):
        raise exception
    elif isinstance(exception, GCSAPIError):
        _gcsapi_error_hook(exception)
    elif isinstance(exception, globus_sdk.GlobusAPIError):
        _default_api_error_hook(exception)
    else:
        click.echo(u"{}: {}".format(exception_type.__name__, exception))
    click.get_current_context().exit(1)


def _default_api_error_hook(exception):
    click.secho("An API Error Occurred", err=True, color="red", bold=True)
    click.secho(
        "HTTP status: {}".format(exception.http_status), err=True, color="yellow"
    )
    click.echo("code:        {}".format(exception.code), err=True)
    if exception.message:
        click.echo("message:     {}".format(exception.message), err=True)


def _auth_timeout_error_hook(exception):
    ids = exception.detail["identities"]
    if len(ids) > 1:
        cmdmsg = "To resolve, you should run one of these commands"
    else:
        cmdmsg = "To resolve, you should run this command"

    click.secho("An authentication timeout occurred", err=True, color="red", bold=True)
    click.echo(
        "This means you would have been authorized based on your identities, "
        "but you haven't logged in recently enough.",
        err=True,
    )
    click.secho(cmdmsg + " to update your authentication:", err=True, color="yellow")
    for n in _try_resolve_ids_to_usernames(ids):
        click.echo("  globus-connect-server session update {}".format(n), err=True)


def _json_api_error_hook(exception):
    response = exception._underlying_response
    if exception.json_data:
        body = exception.json_data
    else:
        body = response.text
    headers = dict(response.headers)
    gcs_headers = {k: v for k, v in headers.items() if k.startswith("x-gcs-")}
    other_headers = {k: v for k, v in headers.items() if k not in gcs_headers}
    click.secho(
        json.dumps(
            {
                "is_error": True,
                "http_status": exception.http_status,
                "code": exception.code,
                "message": exception.message,
                "body": body,
                "gcs_headers": gcs_headers,
                "headers": other_headers,
            },
            indent=2,
            separators=(",", ": "),
            sort_keys=True,
        ),
        color="yellow",
    )


def _gcsapi_error_hook(exception):
    if outformat_is_text():
        if exception.detail_data_type == "authentication_timeout#1.0.0":
            _auth_timeout_error_hook(exception)
        else:
            _default_api_error_hook(exception)
    else:
        _json_api_error_hook(exception)


def _try_resolve_ids_to_usernames(ids):
    """resolve IDs to usernames, falling back to the IDs which were given"""
    try:
        idmap = globus_sdk.IdentityMap(get_client())
        # add all to do batched lookups
        for i in ids:
            idmap.add(i)
        # do the lookups and produce the usernames
        return [idmap[i]["username"] for i in ids]
    # very conservative error handling -- if anything goes wrong, just return the
    # original IDs
    except Exception:
        return ids
