from __future__ import unicode_literals

from globus_sdk.authorizers import (
    AccessTokenAuthorizer,
    ClientCredentialsAuthorizer,
    RefreshTokenAuthorizer,
)
from globus_sdk.base import BaseClient, merge_params, safe_stringify

from gcs_cli.sdk_client.host_header_ssl_adapter import HostHeaderSSLAdapter
from gcs_cli.sdk_client.paging import PaginatorCollection
from gcs_cli.sdk_client.response import (
    GCSAPIError,
    GCSResponse,
    IterableGCSResponse,
    SingletonGCSResponse,
)


class GCSClient(BaseClient):
    """
    Client for the Globus Connect Server API that is run separately by
    each Globus Connect Server instance starting at GCSv5.

    This class provides helper methods for most common resources in the
    GCS API, and the common low-level interface from
    :class:`BaseClient <globus_sdk.base.BaseClient>` of ``get``, ``put``,
    ``post``, ``patch`` and ``delete`` methods,
    which can be used to access any API resource.

    **Parameters**

    ``gcs_fdqn`` (*str*)
    The fully qualified domain name of the GCS instance this client will
    make calls to

    ``authorizer`` (*:class:`GlobusAuthorizer \
                        <globus_sdk.authorizers.base.GlobusAuthorizer>`*)
    An authorizer instance used for calls to the GCS API

    **Paginated Calls**

    Methods which support pagination can be called as paginated or unpaginated methods.
    If the method name is ``foo``, the paginated version is
    ``paginated.foo``. e.g. ``GCSClient.get_storage_gateway_list`` and
    ``GCSClient.paginated.get_storage_gateway_list`` are both valid.

     **Methods**

     *  :py:meth:`.set_address`
     *  :py:meth:`.get_storage_gateway_list`
     *  :py:meth:`.create_storage_gateway`
     *  :py:meth:`.get_storage_gateway`
     *  :py:meth:`.update_storage_gateway`
     *  :py:meth:`.delete_storage_gateway`
     *  :py:meth:`.get_endpoint`
     *  :py:meth:`.update_endpoint`
     *  :py:meth:`.get_collection_list`
     *  :py:meth:`.create_collection`
     *  :py:meth:`.get_collection`
     *  :py:meth:`.update_collection`
     *  :py:meth:`.delete_collection`
     *  :py:meth:`.get_node_list`
     *  :py:meth:`.get_node`
     *  :py:meth:`.update_node`
     *  :py:meth:`.delete_node`
    """

    # disallow basic auth
    allowed_authorizer_types = [
        AccessTokenAuthorizer,
        RefreshTokenAuthorizer,
        ClientCredentialsAuthorizer,
    ]
    error_class = GCSAPIError
    default_response_class = GCSResponse

    def __init__(self, gcs_fqdn, authorizer=None, **kwargs):
        BaseClient.__init__(
            self,
            "gcs",
            base_url="https://{}/api/".format(gcs_fqdn),
            authorizer=authorizer,
            **kwargs
        )
        self.gcs_fqdn = gcs_fqdn
        # setup the HostHeaderSSLAdapter in case explicit host is set
        self._session.mount("https://", HostHeaderSSLAdapter())
        self.paginated = PaginatorCollection(
            self.get_storage_gateway_list,
            self.get_collection_list,
            self.get_node_list,
            self.get_role_list,
        )

    def set_address(self, address):
        """
        Set a host header. With use of HostHeaderSSLAdapter, this allows requests by IP
        to proceed with an explicit hostname.

        Can accept just the address (raw) or a full URL.
        """
        self._headers["Host"] = self.gcs_fqdn
        if address.startswith("https://"):
            self.base_url = address
        else:
            self.base_url = "https://{}/api/".format(address)

    #
    # Storage Gateways
    #

    def get_storage_gateway_list(self, include=None, **params):
        """
        ``GET /storage_gateways``

        ** Parameters **

        ``include`` (*list of str*)
        List of document types to include in the response. If private_policies
        is in the list, then include private storage gateway policies in the
        attached storage_gateways document. This requires an administrator
        role on the Endpoint.
        """
        if include is not None:
            include = list(map(safe_stringify, include))
        merge_params(params, include=include)

        path = self.qjoin_path("storage_gateways")
        return self.get(path, params=params, response_class=IterableGCSResponse)

    def create_storage_gateway(self, data, **params):
        """
        ``POST /storage_gateways``

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Storage Gateway document, it is recommended
        to use the `StorageGatewayDocumment` class to construct this data.
        """
        path = self.qjoin_path("storage_gateways")
        return self.post(path, data, params=params, response_class=SingletonGCSResponse)

    def get_storage_gateway(self, storage_gateway_id, include=None, **params):
        """
        ``GET /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gatweay_id`` (*str*)
        UUID for the Storage Gateway to be gotten

        ``include`` (*list of str*)
        List of document types to include in the response. If private_policies
        is in the list, then include private storage gateway policies in the
        attached storage_gateways document. This requires an administrator
        role on the Endpoint.
        """
        storage_gateway_id = safe_stringify(storage_gateway_id)
        if include is not None:
            include = list(map(safe_stringify, include))
        merge_params(params, include=include)

        path = self.qjoin_path("storage_gateways", storage_gateway_id)
        return self.get(path, params=params, response_class=SingletonGCSResponse)

    def update_storage_gateway(self, storage_gateway_id, data, **params):
        """
        ``PATCH /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gatweay_id`` (*str*)
        UUID for the Storage Gateway to be updated

        ``data`` (*dict*)
        Data in the format of a Storage Gateway document, it is recommended
        to use the `StorageGatewayDocumment` class to construct this data.
        """
        storage_gateway_id = safe_stringify(storage_gateway_id)
        path = self.qjoin_path("storage_gateways", storage_gateway_id)
        return self.patch(path, data, params=params)

    def delete_storage_gateway(self, storage_gateway_id, **params):
        """
        ``DELETE /storage_gateways/<storage_gateway_id>``

        ** Parameters **

        ``storage_gatweay_id`` (*str*)
        UUID for the Storage Gateway to be deleted
        """
        storage_gateway_id = safe_stringify(storage_gateway_id)
        path = self.qjoin_path("storage_gateways", storage_gateway_id)
        return self.delete(path, params=params)

    #
    # Endpoint
    #

    def get_endpoint(self, **params):
        """
        ``GET /endpoint``
        """
        return self.get("/endpoint", params=params, response_class=SingletonGCSResponse)

    def update_endpoint(self, data, **params):
        """
        ``PATCH /endpoint``

        ** Parameters **

        ``data`` (*dict*)
          Data in the format of a partial Endpoint document.
        """
        return self.patch("/endpoint", data, params=params)

    #
    # Collections
    #

    def get_collection_list(self, include=None, **params):
        """
        GET /collections
        """
        path = self.qjoin_path("collections")
        if include is not None:
            include = safe_stringify(include)
        merge_params(params, include=include)
        return self.get(path, params=params, response_class=IterableGCSResponse)

    def create_collection(self, data, **params):
        """
        POST /collections

        ** Parameters **

        ``data`` (*dict*)
          Data in the format of a Collection document, it is recommended to use
          the `CollectionDocumment` class to construct this data.
        """
        path = self.qjoin_path("collections")
        return self.post(path, data, params=params, response_class=SingletonGCSResponse)

    def get_collection(self, collection_id, **params):
        """
        GET /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be gotten
        """
        collection_id = safe_stringify(collection_id)
        path = self.qjoin_path("collections", collection_id)
        return self.get(path, params=params, response_class=SingletonGCSResponse)

    def update_collection(self, collection_id, data, **params):
        """
        PATCH /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be updated

        ``data`` (*dict*)
        Data in the format of a Collection document, it is recommended to use
        the `CollectionDocumment` class to construct this data.
        """
        collection_id = safe_stringify(collection_id)
        path = self.qjoin_path("collections", collection_id)
        return self.patch(path, data, params=params)

    def delete_collection(self, collection_id, **params):
        """
        DELETE /collections/{collection_id}

        ** Parameters **

        ``collection_id`` (*str*)
        UUID for the Collection to be deleted
        """
        collection_id = safe_stringify(collection_id)
        path = self.qjoin_path("collections", collection_id)
        return self.delete(path, params=params)

    #
    # Nodes
    #

    def get_node_list(self, **params):
        """
        GET /nodes
        """
        path = self.qjoin_path("nodes")
        return self.get(path, params=params, response_class=IterableGCSResponse)

    def get_node(self, node_id, **params):
        """
        GET /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to fetch
        """
        node_id = safe_stringify(node_id)
        self.logger.info("GCSClient.get_node({})".format(node_id))
        path = self.qjoin_path("nodes", node_id)
        return self.get(path, params=params, response_class=SingletonGCSResponse)

    def update_node(self, node_id, data, **params):
        """
        PATCH /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to be updated

        ``data`` (*dict*)
          Data in the format of a Node document.
        """
        node_id = safe_stringify(node_id)
        self.logger.info("GCSClient.update_node({}, ...)".format(node_id))
        path = self.qjoin_path("nodes", node_id)
        return self.patch(path, data, params=params)

    def delete_node(self, node_id, **params):
        """
        DELETE /nodes/{node_id}

        ** Parameters **

        ``node_id`` (*str*)
          UUID for the Node to delete
        """
        node_id = safe_stringify(node_id)
        self.logger.info("GCSClient.delete_node({})".format(node_id))
        path = self.qjoin_path("nodes", node_id)
        return self.delete(path, params=params)

    #
    # Roles
    #

    def get_role_list(self, collection_id=None, include=None, **params):
        """
        GET /roles

        ** Paramaters **

        ``collection_id`` (*str*)
        UUID of a Collection. If given then only roles related to that
        Collection are returned, otherwise only Endpoint roles are returned.

        ``include`` (*str*)
        Pass "all_roles" to request all roles all roles relevant to the
        resource instead of only those the caller has on the resource
        """
        if collection_id is not None:
            collection_id = safe_stringify(collection_id)
        if include is not None:
            include = safe_stringify(include)
        merge_params(params, collection_id=collection_id, include=include)

        path = self.qjoin_path("roles")
        return self.get(path, params=params, response_class=IterableGCSResponse)

    def create_role(self, data, **params):
        """
        POST /roles

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Role document, it is recommended to use
        the `RoleDocumment` class to construct this data.
        """
        path = self.qjoin_path("roles")
        return self.post(path, data, params=params, response_class=SingletonGCSResponse)

    def get_role(self, role_id, **params):
        """
        GET /roles/{role_id}

        ** Parameters **

        ``role_id`` (*str*)
        UUID for the Role to be gotten
        """
        role_id = safe_stringify(role_id)
        path = self.qjoin_path("roles", role_id)
        return self.get(path, params=params, response_class=SingletonGCSResponse)

    def delete_role(self, role_id, **params):
        """
        DELETE /roles/{role_id}

        ** Parameters **

        ``role_id`` (*str*)
        UUID for the Role to be deleted
        """
        role_id = safe_stringify(role_id)
        path = self.qjoin_path("roles", role_id)
        return self.delete(path, params=params)
