from typing import Dict, Tuple
import jwt

from flask import (
    current_app,
    g,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
    Blueprint,
)
import marshmallow as ma
from oauth2client import client as google_oauth
from globus.manager.api.models import (
    Endpoint,
    StorageGateway,
    UserCredential,
)
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.utils.update_query_string import update_query_string
import globus.manager.api.errors as errors
from globus.manager.api.blueprints.utils import (
    create_object,
    update_object,
)
from globus.manager.api.schemas.api_v1.request import user_credential_form_post_schema
from sqlalchemy.orm.exc import NoResultFound


class GoogleCloudPlatform:
    __blueprint_name = "google_cloud_platform"
    __connectors = {}  # type: Dict[str, Tuple[str, str]]

    def __init__(self, app):
        blueprint = Blueprint(
            GoogleCloudPlatform.__blueprint_name, __name__, url_prefix="/api/v1"
        )
        blueprint.add_url_rule(
            "/authcallback_google",
            "authcallback_google",
            self.authcallback_google,
            methods=["GET"],
        )
        blueprint.add_url_rule(
            "/user_credentials",
            "user_credentials_post",
            self.user_credentials_post,
            methods=["POST"],
        )

        blueprint.register_error_handler(
            errors.APIBaseError, self.handle_validation_error
        )

        if GoogleCloudPlatform.__blueprint_name not in app.blueprints.keys():
            app.register_blueprint(blueprint)

    def add_connector(self, connector_id, scopes, user_credential_policies_data_type):
        GoogleCloudPlatform.__connectors[connector_id] = (
            scopes,
            user_credential_policies_data_type,
        )

    @authenticated
    def user_credentials_post(self):
        try:
            post_data = user_credential_form_post_schema.load(request.form)
        except ma.ValidationError as e:
            raise errors.ValidationError(exc=e)

        # confirm post data is for a valid storage gateway
        storage_gateway = None
        try:
            storage_gateway = StorageGateway.query.filter_by(
                deleted=False, id=post_data["storage_gateway"]
            ).one()
        except NoResultFound:
            current_app.logger.info(
                "Attempt to create a user credential with an unknown or "
                "disabled storage gateway: {}".format(post_data["storage_gateway"])
            )
            raise errors.BadRequestError(
                "storage_gateway does not exist, is "
                "disabled, or does not allow guest collections"
            )

        if storage_gateway.connector_id not in self.__connectors:
            raise errors.UnsupportedMediaError()

        authorizer = StorageGatewayAuthorizer(storage_gateway)
        try:
            authorized_identity = authorizer.authorize_identity(
                post_data["identity_id"], g.auth_state
            )
        except errors.ForbiddenError as e:
            query_updates = {
                "error": e.error,
                "error_description": e.error_description,
            }
            if e.detail:
                if isinstance(e.detail, errors.NotFromAllowedDomain):
                    query_updates["domains"] = ",".join(e.detail.allowed_domains)
                elif isinstance(e.detail, errors.AuthenticationTimeout):
                    query_updates["identities"] = ",".join(e.detail.identities)
            if "redirect_uri" in post_data:
                redirect_to = update_query_string(
                    post_data["redirect_uri"], query_updates
                )
                return redirect(redirect_to)
            else:
                raise

        post_data["authorized_id_data"] = authorized_identity

        app_secret = Endpoint.query.one().oauth_state_key
        state = jwt.encode(post_data, app_secret, algorithm="HS256")

        # Set up the oauth2 flow.
        # access_type is `offline` by default.
        # response_type is `code` by default.
        # prompt must be set to `consent` to get a refresh token,
        # even when access_type is offline.
        flow = google_oauth.OAuth2WebServerFlow(
            state=state,
            client_id=storage_gateway.client_id,
            client_secret=storage_gateway.secret,
            scope=GoogleCloudPlatform.__connectors[storage_gateway.connector.id][0],
            redirect_uri=url_for(
                ".".join([GoogleCloudPlatform.__blueprint_name, "authcallback_google"]),
                _external=True,
                _scheme="https",
            ),
            prompt="consent",
        )

        mapped_to = storage_gateway.map_identity(authorized_identity)
        split_username = mapped_to.split("@")

        flow.params["login_hint"] = "@".join(split_username[0:2])
        flow.params["hd"] = split_username[1]

        # respond with a redirect
        return redirect(flow.step1_get_authorize_url(), code=303)

    def authcallback_google(self):
        state = request.args.get("state")

        if not state:
            # Every request to this route MUST have a state parameter.
            raise errors.BadRequestError("missing state parameter")

        app_secret = Endpoint.query.one().oauth_state_key

        try:
            state_data = jwt.decode(state, app_secret, algorithms=["HS256"])
        except jwt.exceptions.InvalidTokenError:
            raise errors.BadRequestError("Invalid state parameter")

        error = request.args.get("error")
        if error:
            # Google sent us back with an error. Send the user-agent
            # back to the `redirect_uri` that was provided in the original
            # POST to /user_credentials + the error as an added query param.
            redirect_to = update_query_string(
                state_data["redirect_uri"], dict(error=error)
            )

            current_app.logger.error("authcallback_google error: {}".format(error))

            return redirect(redirect_to)

        try:
            storage_gateway = StorageGateway.query.filter_by(
                deleted=False, id=state_data["storage_gateway"]
            ).one()
        except NoResultFound:
            current_app.logger.error(
                "Made it to /authcallback_google with an unknown "
                "storage_gateway {}.".format(state_data["storage_gateway"])
            )
            raise errors.BadRequestError("storage_gateway does not exist")

        code = request.args.get("code")
        if code:
            # If we have a `code` we can proceed to exchange it for
            # an access token.
            flow = google_oauth.OAuth2WebServerFlow(
                client_id=storage_gateway.client_id,
                client_secret=storage_gateway.secret,
                scope=GoogleCloudPlatform.__connectors[storage_gateway.connector.id][0],
                redirect_uri=url_for(
                    ".".join(
                        [GoogleCloudPlatform.__blueprint_name, "authcallback_google"]
                    ),
                    _external=True,
                    _scheme="https",
                ),
            )

            try:
                credentials = flow.step2_exchange(code)
            except google_oauth.FlowExchangeError as ex:
                # FlowExchangeError will be the `error` and `error_description`
                # fields concantenated together, so you might see something
                # like `invalid_grantCode was already redeemed`.
                current_app.logger.exception("Error during Google authentication flow")
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(error="flow_exchange_error", error_description=ex),
                )

                return redirect(redirect_to)

            validated_email = credentials.id_token["email"]

            authorized_identity = state_data["authorized_id_data"]
            mapped_to = storage_gateway.map_identity(authorized_identity)
            if not mapped_to:
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(
                        error="invalid_user",
                        error_description="Storage gateway cannot map identity",
                    ),
                )
                return redirect(redirect_to)

            if mapped_to != validated_email:
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(
                        error="invalid_user",
                        error_description="Globus identity does not "
                        "map to the requested user",
                    ),
                )
                return redirect(redirect_to)

            if not storage_gateway.check_user_allowed(mapped_to):
                redirect_to = update_query_string(
                    state_data["redirect_uri"],
                    dict(
                        error="invalid_user",
                        error_description="Google identity is not "
                        "allowed by the storage gateway",
                    ),
                )

            user_credential = UserCredential.query.filter_by(
                storage_gateway_id=storage_gateway.id,
                username=validated_email,
                identity_id=state_data["identity_id"],
            ).one_or_none()

            if user_credential:
                update_object(
                    user_credential,
                    lambda o, credentials: o.put_credentials(credentials),
                    credentials,
                )
            else:
                # create new user credential
                model_class = UserCredential.polymorphic_class(
                    storage_gateway.connector_id
                )

                user_credential = model_class(
                    doc={
                        "DATA_TYPE": "user_credential#1.0.0",
                        "identity_id": state_data["identity_id"],
                        "username": mapped_to,
                        "storage_gateway_id": storage_gateway.id,
                    },
                    storage_gateway=storage_gateway,
                    credentials=credentials,
                )
                create_object(user_credential)
            redirect_to = update_query_string(
                state_data["redirect_uri"], dict(user_credential_id=user_credential.id)
            )

            return redirect(redirect_to)

        # if the request has neither an `error` or a `code` it's a bad request.
        raise errors.BadRequestError()

    @staticmethod
    def negotiate_error_response(request, error):
        if "text/html" == request.accept_mimetypes.best:
            response = make_response(
                render_template(
                    "error.jinja2", status_code=error.status_code, **error.to_dict()
                )
            )
        else:
            error_dict = error.to_dict()
            error_dict.pop("detail")
            response = jsonify(error_dict)

        response.status_code = error.status_code

        return response

    @staticmethod
    def handle_validation_error(error):
        return GoogleCloudPlatform.negotiate_error_response(request, error)
