from collections.abc import Mapping
import os
import re
from typing import List

from flask import current_app, render_template
import globus_sdk
from sqlalchemy import orm
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict

from globus.manager.api.utils import virtual_client
from globus.manager.api.utils.log import log_transfer_exception
from globus.manager.utils.run_as_user import run_as_user
from globus.manager.version import __version__

from . import (
    db,
    ConfigurationObject,
    Domain,
    Role,
    StorageGateway,
    UserCredential,
)
from globus.manager.utils.path_restriction import PathRestrictions

from globus.manager.constants import CollectionType, ConfigurationObjectTag
from globus.manager.api.schemas.transfer import (
    mapped_collection_create_schema,
    mapped_collection_update_schema,
    guest_collection_create_schema,
    guest_collection_update_schema,
)
from globus.manager.api.schemas.api.collection import collection_private_schema
import globus.manager.api.errors as errors


class Collection(ConfigurationObject):
    """
    Collection model for mapped and guest collections in GCSv5
    """

    __tablename__ = "collection"

    @classmethod
    def polymorphic_class(cls, discriminator):
        return ConfigurationObject.polymorphic_class(
            ConfigurationObjectTag.COLLECTION.name, discriminator
        )

    _DATA_TYPE = "collection#1.0.0"
    collection_properties = [
        "collection_type",
        "deleted",
        "mapped_collection_id",
        "storage_gateway_id",
        "user_credential_id",
        "domain_name",
        "display_name",
        "identity_id",
        "collection_base_path",
        "default_directory",
        "sharing_restrict_paths",
        "public",
        "force_encryption",
        "disable_verify",
        "allow_guest_collections",
        "organization",
        "department",
        "keywords",
        "description",
        "contact_email",
        "contact_info",
        "info_link",
        "policies",
    ]

    connector_policies = []  # type: List[str]

    configuration_schema = collection_private_schema

    # Column definitions
    id = db.Column(
        db.TEXT,
        db.ForeignKey("configuration_object.id", name="col_id"),
        primary_key=True,
    )
    _domain_name = db.Column("domain", db.TEXT, nullable=False, unique=True)
    _user_credential_id = db.Column(
        "user_credential_id",
        db.TEXT,
        db.ForeignKey("user_credential.id"),
        nullable=True,
    )
    user_credential = db.relationship(
        "UserCredential",
        foreign_keys=[_user_credential_id],
        uselist=False,
        lazy="joined",
    )
    _storage_gateway_id = db.Column(
        "storage_gateway_id",
        db.TEXT,
        db.ForeignKey("storage_gateway.id"),
        nullable=False,
    )
    _collection_type = db.Column(
        "collection_type", db.Enum(CollectionType), nullable=False
    )
    _allow_guest_collections = db.Column(
        "allow_guest_collections", db.BOOLEAN, nullable=False, server_default="0"
    )
    _mapped_collection_id = db.Column(
        "mapped_collection_id", db.TEXT, db.ForeignKey("collection.id"), nullable=True
    )
    storage_gateway = db.relationship(
        "StorageGateway",
        foreign_keys=[_storage_gateway_id],
        uselist=False,
        lazy="joined",
    )
    mapped_collection = db.relationship(
        "ConfigurationObject",
        primaryjoin="remote(ConfigurationObject.id) "
        "== foreign(Collection.mapped_collection_id)",
        uselist=False,
        viewonly=True,
        lazy="joined",
    )
    owner_role = db.relationship(
        "Role",
        primaryjoin="and_(Collection.id==Role.resource, Role.primary_role)",
        uselist=False,
        backref="collection",
    )
    endpoint = db.relationship(
        "Endpoint",
        primaryjoin="remote(Endpoint.id) != foreign(Collection.id)",
        uselist=False,
        viewonly=True,
    )

    # methods and properties
    def __init__(
        self,
        appsync_data=None,
        doc=None,
        storage_gateway=None,
        mapped_collection=None,
        user_credential=None,
        endpoint=None,
    ):
        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}

            args["tag"] = ConfigurationObjectTag.COLLECTION
            args["secondary_discriminator"] = self.connector_id

            if mapped_collection:
                self.mapped_collection = mapped_collection
            if storage_gateway:
                self.storage_gateway = storage_gateway
            if user_credential:
                self.user_credential = user_credential
            if endpoint:
                self.endpoint = endpoint

            super().__init__(**args)
            self.init_on_load()
        elif doc.get("DATA_TYPE") == self._DATA_TYPE:
            self.configuration_hash = MutableDict()
            self._collection_type = doc.get("collection_type")
            self.user_credential_id = doc.get("user_credential_id")
            self.mapped_collection_id = doc.get("mapped_collection_id")
            self.storage_gateway_id = doc.get("storage_gateway_id")

            self.display_name = doc.get("display_name")

            if not doc.get("storage_gateway_id"):
                raise ValueError("Missing required storage_gateway_id")

            if self.is_guest_collection and not doc.get("mapped_collection_id"):
                raise ValueError(
                    "Missing required mapped_collection_id " "for guest collection"
                )

            # Seed relationships on initialization
            if getattr(self, "mapped_collection_id", None):
                if mapped_collection:
                    assert mapped_collection.id == self.mapped_collection_id
                    self.mapped_collection = mapped_collection

            if storage_gateway:
                assert storage_gateway.id == self._storage_gateway_id
                self.storage_gateway = storage_gateway

            if endpoint:
                self.endpoint = endpoint

            if not doc.get("identity_id"):
                raise ValueError("Missing required identity_id")

            self.policies = doc.get("policies", {})

            if self._user_credential_id:
                if not user_credential:
                    raise ValueError("Invalid user_credential_id")
                if self.storage_gateway_id is None:
                    self.storage_gateway_id = user_credential.storage_gateway_id
                elif user_credential.storage_gateway_id != self.storage_gateway_id:
                    raise ValueError("Invalid storage_gateway_id")
            elif self.is_guest_collection and self.storage_gateway.require_credential:
                raise ValueError("Missing required user_credential")

            self.domain_name = doc["domain_name"]
            self.id = self.register_virtual_client()
            id_str = self.id
            super().__init__(
                id=id_str,
                tag=ConfigurationObjectTag.COLLECTION,
                secondary_discriminator=self.connector_id,
            )

            self._put(**doc)
            try:
                admin_role = None
                transfer_doc = None

                transfer_doc = self.create_transfer_endpoint()
                if transfer_doc["id"] != self.id:
                    raise errors.InternalServerError("Bad response from transfer")

                if self.is_guest_collection or current_app.subscription:
                    admin_role = Role(
                        doc=dict(
                            DATA_TYPE="role#1.0.0",
                            collection=self.id,
                            principal="urn:globus:auth:identity:{}".format(
                                self.identity_id
                            ),
                            role="administrator",
                        ),
                        primary_role=True,
                    )
                self.owner_role = admin_role
            except Exception as e:
                current_app.logger.warning("Error creating resource: %s", e)
                if admin_role:
                    admin_role.delete()
                if transfer_doc:
                    self.delete_transfer_endpoint(endpoint_id=transfer_doc["id"])
                self.unregister_virtual_client()
                raise
        else:
            raise ValueError("Unsupported DATA_TYPE")

    @orm.reconstructor
    def init_on_load(self):
        if self.configuration_hash:
            ch = self.configuration_hash
            if ch:
                if self.is_guest_collection:
                    ch.pop("sharing_restrict_paths", None)
                self._put(**ch)

    @property
    def manager_url(self):
        """
        HTTPS URL of the GCS Manager service which manages this collection.
        """
        server_name = Domain.query.one().domain_name
        return "https://{}".format(server_name)

    @property
    def https_url(self):
        """
        URL using the "https" scheme which can be used by web clients to
        access data on this collection.
        """
        if not current_app.subscription:
            return None

        return "https://{}".format(self.domain_name)

    @property
    def tlsftp_url(self):
        """
        URL using the "tslftp" scheme which is used by the Globus Transfer
        service to access data on this collection.
        """
        return "tlsftp://{}:443".format(self.domain_name)

    @hybrid_property
    def collection_type(self) -> str:
        """
        Type of this collection: currently one of "mapped" or "guest". This may
        not change after the collection is created.
        """
        if self._collection_type is not None:
            return CollectionType(self._collection_type).name

    @collection_type.expression  # type: ignore
    def collection_type(self):
        return self._collection_type

    @collection_type.setter  # type: ignore
    def collection_type(self, value):
        new_ct = CollectionType(value).name

        old_ct = self.collection_type
        if not old_ct:
            self._collection_type = new_ct
        elif old_ct != new_ct:
            raise ValueError()

    @property
    def display_name(self):
        """
        The user-friendly name for this collection. This may be a Unicode
        string of at most 128 characters with no newline characters.
        """
        return self._display_name

    @display_name.setter
    def display_name(self, value):
        self._display_name = value

    @hybrid_property
    def storage_gateway_id(self):
        return self._storage_gateway_id

    @storage_gateway_id.setter  # type: ignore
    def storage_gateway_id(self, value):
        old_sgid = self._storage_gateway_id

        if old_sgid is not None and old_sgid != value:
            raise ValueError()

        self._storage_gateway_id = value

    @hybrid_property
    def user_credential_id(self):
        if self._user_credential_id:
            return self._user_credential_id
        else:
            raise AttributeError()

    @user_credential_id.setter  # type: ignore
    def user_credential_id(self, value):
        if self.is_mapped_collection:
            if value:
                raise ValueError()
            else:
                return

        old_ucid = self._user_credential_id

        if old_ucid is not None and old_ucid != value:
            raise ValueError()

        self._user_credential_id = value

    @hybrid_property
    def domain_name(self):
        """
        Fully-qualified domain name of the collection. For a mapped collection,
        This may be updated, but is immutable for guest collections.
        """
        return self._domain_name

    @domain_name.setter  # type: ignore
    def domain_name(self, value):
        old_name = self._domain_name
        if value is None:
            raise ValueError()

        if old_name and old_name == value:
            return

        if old_name is not None and self.is_guest_collection:
            raise ValueError("domain_name may not be changed for guest collections")

        self._domain_name = value

    @property
    def identity_id(self):
        return self._identity_id

    @identity_id.setter
    def identity_id(self, value):
        old_identity_id = getattr(self, "_identity_id", None)

        if old_identity_id:
            if value != old_identity_id:
                raise ValueError()
        else:
            try:
                if self.is_guest_collection:
                    uc = self.user_credential
                    if uc:
                        if uc.identity_id != value:
                            raise ValueError()
            except AttributeError:
                pass

            self._identity_id = value

    @property
    def root_path(self):
        if self.is_guest_collection:
            return re.sub(
                "//*",
                "/",
                "/".join([self.mapped_collection.root_path, self.collection_base_path]),
            )
        else:
            return self._root_path

    @property
    def collection_base_path(self):
        return self._collection_base_path

    @collection_base_path.setter
    def collection_base_path(self, value):
        old_value = getattr(self, "collection_base_path", None)
        if old_value is not None and value != old_value:
            raise ValueError("collection_base_path is immutable")

        self._collection_base_path = value

        if self.is_mapped_collection:
            self._root_path = value
        else:
            for bad in ["$HOME", "~"]:
                if bad in value:
                    raise ValueError(
                        "{} not allowed in guest collection base path".format(bad)
                    )

    @property
    def default_directory(self):
        return self._default_directory

    @default_directory.setter
    def default_directory(self, value):
        if not (self.is_mapped_collection and self.collection_base_path == "/"):
            for bad in ["$HOME", "~"]:
                if bad in value:
                    raise ValueError(
                        "{} not allowed in collection default_directory".format(bad)
                    )
        self._default_directory = value

    @default_directory.deleter
    def default_directory(self):
        self._default_directory = None

    public = ConfigurationObject.basic_property("public", required=True)

    @property
    def high_assurance(self):
        return self.storage_gateway.high_assurance

    @property
    def force_encryption(self):
        value = getattr(self, "_force_encryption", None)
        if value is None:
            value = bool(self.storage_gateway.high_assurance)
        return value

    @force_encryption.setter
    def force_encryption(self, value):
        if self.storage_gateway.high_assurance and not value:
            raise ValueError()
        self._force_encryption = value

    @force_encryption.deleter
    def force_encryption(self):
        self._force_encryption = None

    @property
    def disable_verify(self):
        value = getattr(self, "_disable_verify", None)
        if value is None:
            value = False
        return value

    @disable_verify.setter
    def disable_verify(self, value):
        self._disable_verify = value

    @disable_verify.deleter
    def disable_verify(self):
        self._disable_verify = None

    @hybrid_property
    def mapped_collection_id(self):
        if self.is_mapped_collection:
            raise AttributeError()
        else:
            return self._mapped_collection_id

    @mapped_collection_id.expression  # type: ignore
    def mapped_collection_id(cls):
        return cls._mapped_collection_id

    @mapped_collection_id.setter  # type: ignore
    def mapped_collection_id(self, value):
        if self.is_mapped_collection:
            if value and str(value) != self.id:
                raise ValueError()
            else:
                return

        old_value = getattr(self, "_mapped_collection_id", None)

        if old_value is None:
            self._mapped_collection_id = str(value)
        elif old_value != str(value):
            raise ValueError()

    @mapped_collection_id.deleter  # type: ignore
    def mapped_collection_id(self):
        if self.is_guest_collection:
            raise AttributeError()

    @hybrid_property
    def allow_guest_collections(self):
        if self.is_guest_collection:
            raise AttributeError()

        if not current_app.subscription:
            return False

        return self._allow_guest_collections

    @allow_guest_collections.expression  # type: ignore
    def allow_guest_collections(cls):
        return cls._allow_guest_collections

    @allow_guest_collections.setter  # type: ignore
    def allow_guest_collections(self, value):
        if self.is_guest_collection:
            if value:
                raise ValueError()
            else:
                self._allow_guest_collections = False
        elif not isinstance(value, bool):
            raise ValueError()
        else:
            self._allow_guest_collections = value

    @allow_guest_collections.deleter  # type: ignore
    def allow_guest_collections(self):
        if self.is_mapped_collection:
            self._allow_guest_collections = False

    @property
    def sharing_restrict_paths(self):
        if self.is_guest_collection:
            mcrp = self.mapped_collection.sharing_restrict_paths

            root_path = self.mapped_collection.root_path
            value = PathRestrictions(
                read=[re.sub("//*", "/", "/".join([root_path, p])) for p in mcrp.read],
                read_write=[
                    re.sub("//*", "/", "/".join([root_path, p]))
                    for p in mcrp.read_write
                ],
                none=[re.sub("//*", "/", "/".join([root_path, p])) for p in mcrp.none],
            )
        else:
            value = getattr(self, "_sharing_restrict_paths", None)

        if value is None:
            raise AttributeError()

        return value

    @sharing_restrict_paths.setter
    def sharing_restrict_paths(self, value):
        if self.is_guest_collection:
            raise AttributeError()

        if isinstance(value, Mapping):
            self._sharing_restrict_paths = PathRestrictions(**value)
        elif isinstance(value, PathRestrictions):
            self._sharing_restrict_paths = PathRestrictions(
                DATA_TYPE=value._DATA_TYPE,
                read=value.read,
                read_write=value.read_write,
                none=value.none,
            )
        else:
            raise ValueError()

        all_path_restrictions = (
            self._sharing_restrict_paths.read
            + self._sharing_restrict_paths.read_write
            + self._sharing_restrict_paths.none
        )

        for pr in all_path_restrictions:
            for bad in ["$HOME", "~"]:
                if bad in pr:
                    raise ValueError(
                        "{} not allowed in sharing_restrict_paths".format(bad)
                    )

    @sharing_restrict_paths.deleter
    def sharing_restrict_paths(self):
        self._sharing_restrict_paths = None

    def absolute_root_path(self, username):
        """
        Call :py:code:`StorageGateway.replace_path_vars()` on the root
        path, replacing $HOME, ~, and $USER.

        :param username:
            Connector-specific username to use for $USER.
        :returns:
            A PathRestrictions object or None.
        """
        return StorageGateway.replace_path_vars(
            self.root_path,
            username=username,
            home_directory=self.storage_gateway.home_directory(username),
        )

    def absolute_sharing_restrict_paths(self, username):
        """
        Call :py:code:`StorageGateway.replace_path_vars()` on each path in the
        sharing path restrictions for this collection and return a new
        PathRestrictions object with the new paths.

        :param username:
            Connector-specific username to use for $USER.
        :returns:
            A PathRestrictions object or None.
        """
        sharing_restrict_paths = getattr(self, "sharing_restrict_paths", None)
        if sharing_restrict_paths:
            home_directory = self.storage_gateway.home_directory(username)

            absolute_paths = PathRestrictions()
            absolute_paths.read = [
                StorageGateway.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (sharing_restrict_paths.read or [])
            ]
            absolute_paths.read_write = [
                StorageGateway.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (sharing_restrict_paths.read_write or [])
            ]
            absolute_paths.none = [
                StorageGateway.replace_path_vars(
                    r, username=username, home_directory=home_directory
                )
                for r in (sharing_restrict_paths.none or [])
            ]
            return absolute_paths

    organization = ConfigurationObject.basic_property("organization")
    department = ConfigurationObject.basic_property("department")
    keywords = ConfigurationObject.basic_property("keywords")
    description = ConfigurationObject.basic_property("description")
    contact_email = ConfigurationObject.basic_property(
        "contact_email", doc="Email address of the support contact for the collection."
    )
    contact_info = ConfigurationObject.basic_property("contact_info")
    info_link = ConfigurationObject.basic_property("info_link")
    scopes = ConfigurationObject.basic_property("scopes")

    @property
    def gcs_private_uuid(self):
        return self.id

    @property
    def gcs_private_parent_endpoint_id(self):
        return self.endpoint.endpoint_id

    @property
    def host_endpoint(self):
        return self.endpoint.endpoint_id

    @property
    def authentication_timeout_mins(self):
        """
        Authentication timeout in minutes for either HA or non-HA
        collections
        """
        return self.storage_gateway.authentication_timeout_mins

    @staticmethod
    def from_collections_post(
        storage_gateway=None,
        mapped_collection=None,
        domain_name=None,
        endpoint=None,
        user_credential=None,
        authorized_identity=None,
        doc=None,
    ):

        if doc is None or doc.get("DATA_TYPE") != "collection#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")

        force_encryption = doc.get("force_encryption")
        if storage_gateway.high_assurance:
            force_encryption = True

        if doc["collection_type"] == CollectionType.mapped:
            model_class = Collection.polymorphic_class(storage_gateway.connector_id)

            policies = doc.get("policies", {})

            if policies:
                policies_data_type = policies.get("DATA_TYPE")
                if policies_data_type != model_class.POLICIES_DATA_TYPE:
                    raise ValueError("Unsupported DATA_TYPE")

            return model_class(
                doc={
                    **doc,
                    **{
                        "storage_gateway_id": storage_gateway.id,
                        "identity_id": authorized_identity["id"],
                        "domain_name": domain_name,
                        "force_encryption": force_encryption,
                    },
                },
                storage_gateway=storage_gateway,
                endpoint=endpoint,
            )
        else:
            account = storage_gateway.map_identity(authorized_identity)
            allowed = storage_gateway.check_user_allowed(account)
            if not allowed:
                raise errors.ForbiddenError()

            if storage_gateway.require_credential:
                if user_credential is None:
                    user_credential = UserCredential.query.filter(
                        ~UserCredential.deleted,
                        UserCredential.storage_gateway_id == storage_gateway.id,
                        UserCredential.identity_id == authorized_identity["id"],
                        UserCredential.username == account,
                    ).one_or_none()
                if user_credential is None:
                    raise errors.BadRequestError(
                        "A User Credential is needed to " "create this collection"
                    )

            model_class = Collection.polymorphic_class(storage_gateway.connector_id)

            return model_class(
                doc={
                    **doc,
                    **{
                        "storage_gateway_id": storage_gateway.id,
                        "user_credential_id": user_credential.id
                        if user_credential
                        else None,
                        "identity_id": authorized_identity["id"],
                        "domain_name": domain_name,
                        "force_encryption": force_encryption,
                    },
                },
                storage_gateway=storage_gateway,
                mapped_collection=mapped_collection,
                user_credential=user_credential,
                endpoint=endpoint,
            )

    def register_virtual_client(self):
        """
        register a virtual client for the collection,
        set the collection's id to the virtual client id
        and set the collections scopes
        """
        client, scopes = virtual_client.register(self)

        return client["id"]

    def update_virtual_client(self):
        """
        Update the virtual client. Call this when the domain name
        changes.
        """
        return virtual_client.update_virtual_client(self)

    def unregister_virtual_client(self):
        """
        unregister the virtual client for the collection
        """
        try:
            virtual_client.unregister(self.id)
        except globus_sdk.exc.GlobusAPIError as e:
            if e.http_status != 404:
                raise

    def create_transfer_endpoint(self):
        if self.is_mapped_collection:
            create_ep_doc = mapped_collection_create_schema.dump(self)
            try:
                return current_app.transfer_client.create_endpoint(create_ep_doc)
            except globus_sdk.TransferAPIError as e:
                log_transfer_exception(e)
                raise
        else:
            create_ep_doc = guest_collection_create_schema.dump(self)
            try:
                return current_app.transfer_client.create_shared_endpoint(create_ep_doc)
            except globus_sdk.TransferAPIError as e:
                log_transfer_exception(e)
                raise

    def update_scopes(self):
        current_app.logger.debug("Updating scopes for %s", self)
        data_access_scopes = {
            self.id: None,
            self.domain_name: None,
        }

        auth_client = current_app.auth_client
        if self.is_mapped_collection:
            # Mapped collections have a data_access_scope, which includes text
            # related to the dns name, collection name, and allowed_domains. When
            # any of those are updated, we need to update that scope
            params = {
                "scope_strings": ",".join(
                    [
                        "https://auth.globus.org/scopes/{}/data_access".format(midfix)
                        for midfix in [self.id, self.domain_name]
                    ]
                )
            }
            scopes = auth_client.get("/v2/api/scopes", params=params)
            if not scopes.data["scopes"]:
                # Missing, maybe from an old version of GCS?
                resource = "{}/{}/scopes".format(
                    virtual_client.CLIENTS_API_BASE_URL, self.id
                )
                data_access_res = auth_client.post(
                    resource, virtual_client._data_access_scope(self, self.id)
                )
                for data_access_scope in data_access_res.data["scopes"]:
                    scope_parts = data_access_scope["scope_string"].rsplit("/")[-2:]
                    if scope_parts[1] == "data_access":
                        data_access_scopes[scope_parts[0]] = data_access_scope["id"]
            else:
                for scope in scopes.data["scopes"]:
                    current_app.logger.debug("Updating scope %s", scope["id"])
                    updated_scope = virtual_client._data_access_scope(self, self.id)
                    current_app.logger.debug("Updating scope to %s", updated_scope)
                    data_access_res = auth_client.put(
                        "/v2/api/scopes/" + scope["id"], updated_scope
                    )
                    data_access_scope = data_access_res.data["scope"]
                    scope_parts = data_access_scope["scope_string"].rsplit("/")[-2:]
                    if scope_parts[1] == "data_access":
                        data_access_scopes[scope_parts[0]] = data_access_scope["id"]

        # The https scope is used by the webapp to access the collection via
        # https. For a mapped collection, This includes collection name and
        # domain in descriptive text
        params = {
            "scope_strings": ",".join(
                [
                    "https://auth.globus.org/scopes/{}/https".format(midfix)
                    for midfix in [self.id, self.domain_name]
                ]
            )
        }
        scopes = auth_client.get("/v2/api/scopes", params=params)
        for https_scope in scopes.data["scopes"]:
            scope_parts = https_scope["scope_string"].rsplit("/")[-2:]
            updated_scope = virtual_client._https_scope(self, self.id)
            auth_client.put("/v2/api/scopes/" + https_scope["id"], updated_scope)

    def update_transfer_endpoint(self):
        if self.is_mapped_collection:
            ep_doc = mapped_collection_update_schema.dump(self)
        else:
            ep_doc = guest_collection_update_schema.dump(self)

        try:
            current_app.logger.debug("updating endpoint with %s", ep_doc)
            return current_app.transfer_client.update_endpoint(self.id, ep_doc)
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)
            if e.http_status == 404:
                self.deleted = True
            else:
                raise

    def delete_transfer_endpoint(self, endpoint_id=None):
        if endpoint_id is None:
            endpoint_id = self.id
        try:
            res = current_app.transfer_client.delete_endpoint(endpoint_id)
            return res
        except globus_sdk.exc.GlobusAPIError as e:
            log_transfer_exception(e)
            if e.http_status != 404:
                raise

    @run_as_user(user="gcsweb")
    def setup_vhost(self):
        manager_client_secret = current_app.config["CLIENT_SECRET"]
        environment = current_app.config["ENVIRONMENT"]

        if environment == "default":
            auth_endpoint = "https://auth.globus.org"
            transfer_endpoint = "https://transfer.api.globus.org"
        elif environment == "preview":
            auth_endpoint = "https://auth.preview.globus.org"
            transfer_endpoint = "https://transfer.api.preview.globus.org"
        else:
            auth_endpoint = "https://auth.{environment}.globuscs.info".format(
                environment=environment
            )
            transfer_endpoint = "https://transfer.{environment}.api.globusonline.org".format(
                environment=environment
            )

        vhost_conf_dir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")

        vhost_conf_file = os.path.join(vhost_conf_dir, "{}.conf".format(self.id))

        current_app.logger.debug("Writing config to {}".format(vhost_conf_file))

        if not os.path.exists(vhost_conf_dir):
            os.makedirs(vhost_conf_dir, 0o700)

        with open(
            os.open(vhost_conf_file, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, 0o600), "w"
        ) as f:
            f.write(
                render_template(
                    "vhost.conf.jinja2",
                    allowed_domains=self.storage_gateway.allowed_domains,
                    auth_endpoint=auth_endpoint,
                    authentication_timeout_mins=(
                        self.storage_gateway.authentication_timeout_mins
                    ),
                    client_id=self.id,
                    collection_flags=(
                        "RequireHighAssurance"
                        if self.storage_gateway.high_assurance
                        else None
                    ),
                    collection_type=self.collection_type,
                    config_version=__version__,
                    domain_name=self.domain_name,
                    endpoint_domain_name=Domain.query.one().domain_name,
                    secret=manager_client_secret,
                    subscription_id=current_app.subscription,
                    transfer_endpoint=transfer_endpoint,
                    state_key_password=self.endpoint.oauth_state_key,
                    collection=self,
                )
            )

    @run_as_user(user="gcsweb")
    def cleanup_vhost(self):
        vhost_conf_dir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")

        vhost_conf_file = os.path.join(vhost_conf_dir, "{}.conf".format(self.id))

        if os.path.exists(vhost_conf_file):
            os.remove(vhost_conf_file)

    def write_config(self, delete=False):
        if delete:
            self.cleanup_vhost()
        else:
            self.setup_vhost()

    def put(self, **doc):
        old_domain = self.domain_name
        old_display_name = self.display_name
        self._put(**doc)
        self.update_transfer_endpoint()
        if old_domain != self.domain_name or old_display_name != self.display_name:
            # Needed to update the callback uri and name to match the
            # new domain name
            self.update_virtual_client()
            self.update_scopes()

    def _put(self, DATA_TYPE=None, **doc):
        if DATA_TYPE != self._DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        for p in self.collection_properties:
            v = doc.get(p)
            if v is not None:
                setattr(self, p, v)
            elif hasattr(self, p) and p not in ["root_path", "deleted"]:
                delattr(self, p)
        self.policies = doc.get("policies", None)
        self.refresh_configuration_hash()

    @property
    def policies(self):
        return self.configuration_schema.dump_policies(self)

    @policies.setter
    def policies(self, value):
        if value is None:
            value = {}

        # Verify everything in policies is relevant to this connector
        for p in value:
            if p == "DATA_TYPE":
                if value[p] != self.POLICIES_DATA_TYPE:
                    raise ValueError()
                continue

            if p not in self.connector_policies:
                raise ValueError("Unknown policy {}".format(p))

        # Update policies
        for p in self.connector_policies:
            v = value.get(p)

            if v is not None:
                setattr(self, p, v)
            elif hasattr(self, p):
                delattr(self, p)
            else:
                setattr(self, p, None)

    @policies.deleter
    def policies(self):
        for p in self.connector_policies:
            delattr(self, p)

    def patch(self, **doc):
        """
        Apply PATCH semantics to this Collection.
        """
        old_domain = self.domain_name
        old_display_name = self.display_name
        self._patch(**doc)
        self.update_transfer_endpoint()
        if old_domain != self.domain_name or old_display_name != self.display_name:
            # Needed to update the callback uri and name to match the
            # new domain name
            self.update_virtual_client()
            self.update_scopes()

    def _patch(self, DATA_TYPE=None, policies=None, **doc):
        if DATA_TYPE != self._DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        # Verify everything in doc is writable
        for p in doc:
            if p == "policies":
                continue
            if p not in self.collection_properties:
                raise ValueError("Unsupported property {}".format(p))

        # Update properties
        for p in self.collection_properties:
            if p in doc:
                v = doc[p]
                current_app.logger.debug(
                    "Patching %s from %s to %s", p, getattr(self, p, None), v
                )
                if v is not None:
                    setattr(self, p, v)
                elif hasattr(self, p):
                    delattr(self, p)

        if policies:
            # Verify everything in policies is understood by connector
            for p in policies:
                if p == "DATA_TYPE":
                    continue

                if p not in self.connector_policies:
                    raise ValueError("Unknown policy {}".format(p))

            # Update policies
            for p in self.connector_policies:
                v = policies.get(p)
                if v is not None:
                    setattr(self, p, v)
                elif hasattr(self, p):
                    delattr(self, p)
                else:
                    setattr(self, p, None)
        self.refresh_configuration_hash()

    def delete(self):
        if self.is_mapped_collection or not self.mapped_collection.deleted:
            self.delete_transfer_endpoint()
        if self.owner_role:
            self.owner_role.primary_role = False
        self.unregister_virtual_client()
        super().delete()

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash,
        including updating transfer.
        """
        if data:
            self.update(data["items"][0])
        self.init_on_load()
        self.update_transfer_endpoint()
        self.refresh_configuration_hash()

    def update(self, new_state):
        super().update(new_state)

    def check_user_credential(self, credential):
        """
        Connector implementations can check whether this credential can be used
        to access this collection. If not, they must raise an appropriate error,
        and if appropriate mark the credential as invalid (but not add it to
        the session).
        """
        pass

    @hybrid_property
    def is_mapped_collection(self):
        return self.collection_type == CollectionType.mapped.name

    @is_mapped_collection.expression  # type: ignore
    def is_mapped_collection(self):
        return self._collection_type == CollectionType.mapped.name

    @hybrid_property
    def is_guest_collection(self):
        return self.collection_type == CollectionType.guest.name

    @is_guest_collection.expression  # type: ignore
    def is_guest_collection(self):
        return self._collection_type == CollectionType.guest.name

    def __repr__(self):
        return "<Collection {}>".format(self.id)
