import uuid

from flask import current_app

from . import db
from .base import Base
from globus.manager.constants import ConfigurationObjectTag

from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import validates
from sqlalchemy.types import TypeDecorator


class EncryptedConfigurationType(TypeDecorator):
    """
    SQLAlchemy Type decorator that encrypts a column's value with the current
    flask app's keychain.
    """

    impl = db.TEXT

    def process_bind_param(self, value, dialect):
        if value:
            value = current_app.keychain.encrypt(value)
        else:
            value = "{}"
        return value

    def process_result_value(self, value, dialect):
        if value and value != "{}":
            value = MutableDict(current_app.keychain.decrypt(value))
        else:
            value = MutableDict()
        return value


class ConfigurationObject(Base):
    """
    ConfigurationObject

    The ConfigurationObject model is the parent of all models which are
    stored in an encrypted state in the GCS AppSync database. The table
    here mirrors the contents of the appsync database. Subclasses which
    are stored in the database must implement the following:

    - configuration_schema
        An object with a "dump" method that can process the model into a JSON
        object. This is used to update the configuration_hash column when the
        object is flushed to the database.
    - constructor
        The constructor for this object takes parameters for its columns:
        id, tag, serial, configuration_hash, and
        deleted. By default, this constructor uses the "tag" value (a
        ConfigurationObjectTag) as a discriminator for SQLalchemy's
        polymorphic object loading. An additional value may be passed as
        they secondary_discriminator parameter for objects where the tag plus
        some other data (most often a connector uuid) is needed to determine
        which class to use to load an object from the database.

        By convention, each subclass constructor accepts arguments of either
        appsync_data, or doc. appsync_data is used for data received
        from the appsync service that has already been validated by a DTN.
        It's a dict that contains the following values:

        -   owner_id: auth client id used to create this endpoint
        -   id: object id
        -   tag: one of the globus.manager.constants.ConfigurationObjectTag
            enum values
        -   serial: integer used to detect change conflicts--each change to
            the object in appsync increments this value.
        -   created_at: the time apspync created the object
        -   updated_at: the time apspync last updated the object
        -   configuration: encrypted data containing the object state. If this
            is "{}". there is no state yet or the object has been deleted.
        -   deleted: a boolean indicating that the object has been deleted in
            appsync.

        The doc argument is used when passing some sort of document containing
        the definition of the object. In many cases, this matches the values in
        the API schema, but may be different in some cases, based on the
        representation of the object. This is typically PUT into the object to
        set its state via properties, and then those proeprties are serialized
        to configuration when the object is prepared to be written to the
        database.

    Data encryption:

    The configuration_hash column stores a dict of data in the database. This
    data is encrypted at rest in the local database, using methods in the
    EncryptedConfigurationType, so properties that are stored in that are not
    easily queryable. For those cases, if the data needs to be queryied, use
    additional columns and indices in the object-type specific tables to handle
    that data.
    """

    __tablename__ = "configuration_object"
    __mapper_args__ = {"polymorphic_on": "discriminator"}

    # Database Columns
    id = db.Column("id", db.TEXT, primary_key=True, doc="ID of this object")
    tag = db.Column(
        "tag",
        db.Enum(ConfigurationObjectTag),
        nullable=False,
        doc="""
        The ConfigurationObjectTag used when sending this object to
        appsync. This allows us to get a rough idea of what is being stored
        there, without revealing too much about what the contents are.
        """,
    )
    serial = db.Column(
        db.INTEGER,
        nullable=False,
        doc="""
        The version number of this object. When an object is updated in
        appsync, this is checked to ensure that the new version is a change
        based on the previous version, and if not a conflict error is
        returned
        """,
    )
    discriminator = db.Column(
        "discriminator",
        db.TEXT,
        nullable=False,
        default="",
        server_default="",
        doc="""
        A string that allows SQLAlchemy determine which class to use when
        loading the data into the object. This is typically either the tag
        value or the tag with the UUID of the related connector.
        """,
    )
    configuration_hash = db.Column(
        "configuration",
        MutableDict.as_mutable(EncryptedConfigurationType),
        nullable=False,
        doc="""
        A dict containing the json document for this object. This is
        stored in the database as an encrypted binary blob in the
        "configuration" column.
        """,
    )
    _deleted = db.Column(
        "deleted",
        db.BOOLEAN,
        server_default="0",
        nullable=False,
        doc="""
        A flag indicating that an object has been soft deleted. In that case,
        the object's configuration may be the string `{}` to indicate that no
        data is available for the object any more.
        """,
    )

    # Static methods to manage some common patters we'll use in subclasses
    @staticmethod
    def basic_property(name, required=False, doc=None):
        """
        Create a propery on an class allowing setting, getting, and deleting.
        The low-level SQLAlchemy object constructor will only set kwargs
        passed to it if there are properties defined by the class or in
        __init__. Use this as a replacement for using @property to declare
        properties without any special checking and which are serialized into
        the configuration_hash but not otherwise represented in a database
        column.
        """
        _name = "_" + name

        def fget(self):
            v = getattr(self, _name)
            if v is None:
                raise AttributeError()
            return v

        def fset(self, value):
            if required and value is None:
                raise ValueError()
            setattr(self, _name, value)

        def fdel(self):
            if required:
                raise AttributeError()
            setattr(self, _name, None)

        return property(fset=fset, fget=fget, fdel=fdel, doc=doc)

    @classmethod
    def polymorphic_class(cls, discriminator, secondary=""):
        """
        Determines which model subclass constructor to use when given an
        encrypted blob, based on some discriminator value.

        TODO: currently, the discriminator is obtained by concatenating the TAG
        value which is stored in AppSync with an optional secondary
        discriminator. The secondary discriminator is used to choose when
        multiple subclasses implement different versions of the model. However,
        the value of the discriminator is not currently stored in the appsync
        schema, so is not available. It might be simpler to include a new field
        in the appsync schema and use that (uuid?) to choose which constructor
        to use to determine the class.
        """
        # Is there API for this?
        mapped = cls.__mapper__.polymorphic_map.get(
            ConfigurationObjectTag(discriminator).name + secondary
        )
        if mapped:
            return mapped.class_

    # Methods and properties for this class
    def __init__(self, id=None, tag=None, **kwargs):

        assert isinstance(tag, ConfigurationObjectTag)
        tag_string = tag.name

        secondary_discriminator = kwargs.pop("secondary_discriminator", "")

        if self.configuration_hash is None:
            self.configuration_hash = MutableDict()

        if "configuration_hash" in kwargs:
            kwargs["configuration_hash"] = MutableDict(**kwargs["configuration_hash"])

        super().__init__(
            id=id, tag=tag, discriminator=tag_string + secondary_discriminator, **kwargs
        )

        if self.id is None:
            if self.configuration_hash.get("id"):
                self.id = self.configuration_hash["id"]
            else:
                self.id = str(uuid.uuid4())
        if self.serial is None:
            self.serial = 0

    @validates("id")
    def immutable_properties(self, key, value):
        """
        Ensure that the "id" property is only ever set once on this object.
        """
        old_value = getattr(self, key, None)
        if isinstance(value, uuid.UUID):
            value_str = str(value)
        else:
            value_str = value
        if old_value is not None and old_value != value_str:
            # TODO: specialize the error for violation?
            raise ValueError("The {} property is immutable".format(key))
        return value_str

    def update(self, new_state):
        assert self.tag == ConfigurationObjectTag(new_state["tag"])
        self.created_at = new_state["created_at"]
        self.updated_at = new_state["updated_at"]
        self.serial = new_state["serial"]
        self.configuration = new_state["configuration"]
        self.deleted = new_state["deleted"]
        if self.configuration_hash.get("deleted"):
            self.deleted = True

        if hasattr(self, "init_on_load"):
            self.init_on_load()

    def delete(self):
        self.deleted = True
        self.refresh_configuration_hash()

    @property
    def configuration(self):
        """
        A binary blob containing the encrypted data from appsync. This data
        can only be encrypt4ed by retrieving the DataKey object from
        appsync.
        """
        self.refresh_configuration_hash()

        if self.configuration_hash:
            return current_app.keychain.encrypt(self.configuration_hash)
        else:
            return "{}"

    @configuration.setter  # type: ignore
    def configuration(self, value):
        if value == "{}":
            self.configuration_hash = MutableDict()
        else:
            self.configuration_hash = MutableDict(**current_app.keychain.decrypt(value))
            if self.configuration_hash and self.configuration_hash.get("deleted"):
                self.deleted = True

    @hybrid_property
    def deleted(self):
        return bool(self._deleted)

    @deleted.expression  # type: ignore
    def deleted(cls):
        return cls._deleted

    @deleted.setter  # type: ignore
    def deleted(self, value):
        if value:
            self._deleted = value
        else:
            self._deleted = False

    def write_config(self, delete: bool = False):
        """
        Models override this to have config files persisted to the filesystem
        for other services to find.

        :param delete:
            If True, this model is being removed from the system and its local
            filesystem config should also be deleted.
        """
        pass

    def refresh_configuration_hash(self):
        new_hash = self.configuration_schema.dump(self)
        if new_hash != self.configuration_hash:
            self.configuration_hash = MutableDict(**new_hash)

    def __repr__(self):
        return "<ConfigurationObject {} {}>".format(self.tag.name, self.id)
