from .collection import Collection
from globus.manager.constants import ConfigurationObjectTag, CollectionType


class GoogleCloudStorageCollection(Collection):
    POLICIES_DATA_TYPE = "google_cloud_storage_collection_policies#1.0.0"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
        )
    }

    def __init__(
        self,
        appsync_data=None,
        doc=None,
        storage_gateway=None,
        mapped_collection=None,
        user_credential=None,
        endpoint=None,
    ):
        if doc:
            policies = doc.get("policies")
            if not policies:
                policies = {
                    "DATA_TYPE": self.POLICIES_DATA_TYPE,
                }

            project = policies.get("project")
            collection_type = CollectionType(doc.get("collection_type"))

            if collection_type == CollectionType.guest:
                if project is None:
                    project = mapped_collection.project
                    policies = {**policies, **{"project": project}}
                if not user_credential:
                    raise ValueError("user_credential")

                # Raises exception if not possible
                user_credential.can_access_project(project)
            elif project is None:
                # If the storage gateway defines only one project, use it as
                # the default if none is specified
                allowed_projects = getattr(storage_gateway, "projects", [])
                if len(allowed_projects) == 1:
                    project = allowed_projects[0]
                    policies = {**policies, **{"project": project}}

            doc = {**doc, "policies": {**policies}}

        super().__init__(
            appsync_data=appsync_data,
            doc=doc,
            storage_gateway=storage_gateway,
            mapped_collection=mapped_collection,
            user_credential=user_credential,
            endpoint=endpoint,
        )

    @property
    def connector_id(self):
        """
        UUID of the Google Cloud Storage connector
        """
        return "56366b96-ac98-11e9-abac-9cb6d0d9fd63"

    @connector_id.setter
    def connector_id(self, value):
        if value and str(value) != self.connector_id:
            raise ValueError("Immutable property connector_id")

    connector_policies = ["project"]

    @property
    def project(self):
        """
        Project associated with this collection. All users of this collection
        must have Google credentials which are authorized to use this account,
        and all access to Google Cloud Storage buckets will be charged to this
        account.
        """
        if self.is_guest_collection:
            return self.mapped_collection.project
        else:
            return self._project

    @project.setter
    def project(self, value):
        if not value:
            raise ValueError("Google Cloud Storage guest collections require a project")

        allowed_projects = getattr(self.storage_gateway, "projects", [])
        if allowed_projects and value not in allowed_projects:
            raise ValueError()

        if self.is_guest_collection and value != self.mapped_collection.project:
            raise ValueError("project")

        self._project = value

    def check_user_credential(self, credential):
        credential.can_access_project(self.project)

    def __repr__(self):
        return "<GoogleCloudStorageCollection {}>".format(self.id)
