from flask import current_app
import marshmallow as ma

from .connector import Connector
from .domain import Domain
from .storage_gateway import StorageGateway

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag


class GoogleDriveStorageGateway(StorageGateway):
    CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }
    connector_policy_names = [
        "client_id",
        "secret",
        "user_api_rate_quota",
    ]

    connector = Connector(
        id=CONNECTOR_ID, display_name="Google Drive", can_post_credential=False
    )

    def __init__(self, appsync_data=None, doc=None):
        super().__init__(appsync_data=appsync_data, doc=doc)

        if doc:
            missing = []

            if not self.client_id:
                missing.append("client_id")

            if not self.secret:
                missing.append("secret")

            if missing:
                raise errors.ValidationError(
                    "Missing required value(s) in policies document: {}".format(
                        ", ".join(missing)
                    ),
                    exc=ma.ValidationError(
                        {prop: "Missing data for required field" for prop in missing}
                    ),
                )

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError("Invalid client_id")
        self._client_id = value

    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError("Invalid secret")
        self._secret = value

    @property
    def user_api_rate_quota(self):
        if self._user_api_rate_quota is None:
            raise AttributeError()

        return self._user_api_rate_quota

    @user_api_rate_quota.setter
    def user_api_rate_quota(self, value):
        if value is not None and not isinstance(value, int):
            raise ValueError("Invalid user_api_rate_quota")
        self._user_api_rate_quota = value
        self.configuration_hash.changed()

    @user_api_rate_quota.deleter
    def user_api_rate_quota(self):
        self._user_api_rate_quota = None

    def check_user_allowed(self, username):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    @property
    def load_dsi_module(self):
        return "google_drive"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        if value is not None and value != self.load_dsi_module:
            raise ValueError("Invalid load_dsi_module")

    def home_directory(self, username):
        return "/My Drive"

    def delete(self):
        self._client_id = None
        self._secret = None
        super().delete()

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """
        endpoint_domain = Domain.query.one().domain_name
        return "https://{}/api/v1/authcallback_google".format(endpoint_domain)

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        if value is not None and value != self.process_user:
            raise ValueError("Invalid process_user")

    def put_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_drive_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")

        super().put_policies(**policies)

    def patch_policies(self, DATA_TYPE=None, **policies):
        if DATA_TYPE != "google_drive_storage_policies#1.0.0":
            raise ValueError("Unsupported DATA_TYPE")
        super().patch_policies(**policies)
