from globus.manager.constants import ConfigurationObjectTag

from .user_credential import UserCredential


class S3UserCredential(UserCredential):
    POLICY_DATA_TYPE = "s3_user_credential_policies#1.0.0"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name
            + "7643e831-5f6c-4b47-a07f-8ee90f401d23"
        )
    }
    connector_properties = ["s3_key_id", "s3_secret_key"]
    anonymous_display_name = "Anonymous S3"
    unconfigured_display_name = "Unprovisioned S3 Credential"

    def __init__(self, appsync_data=None, doc=None, storage_gateway=None):
        if doc:
            invalid = False
            provisioned = False
            policies = doc.get("policies", {})
            s3_key_id = policies.get("s3_key_id")
            s3_secret_key = policies.get("s3_secret_key")

            if storage_gateway.s3_user_credential_required:
                if s3_key_id:
                    display_name = s3_key_id
                else:
                    display_name = self.unconfigured_display_name
                if not policies:
                    policies = {
                        "DATA_TYPE": self.POLICY_DATA_TYPE,
                    }
                if not (s3_key_id and s3_secret_key):
                    invalid = True
                else:
                    provisioned = True
            else:
                display_name = self.anonymous_display_name
                if s3_key_id or s3_secret_key:
                    raise ValueError("key not allowed for anonymous access")
                invalid = False
                provisioned = True
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }
            doc = {
                **doc,
                **{
                    "display_name": display_name,
                    "invalid": invalid,
                    "deleted": False,
                    "provisioned": provisioned,
                    "policies": policies,
                },
            }
        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

    @property
    def connector_id(self):
        return "7643e831-5f6c-4b47-a07f-8ee90f401d23"

    @connector_id.setter
    def connector_id(self, value):
        if value and str(value) != self.connector_id:
            raise ValueError("Immutable property connector_id")

    @property
    def s3_key_id(self):
        value = self._s3_key_id
        if value is None:
            raise AttributeError()
        return value

    @s3_key_id.setter
    def s3_key_id(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self._s3_key_id = value
            if value is None:
                self.invalid = True
                self.display_name = "Unprovisioned S3 Credential"
        elif value:
            raise ValueError()

    @s3_key_id.deleter
    def s3_key_id(self):
        if self.storage_gateway.s3_user_credential_required:
            self._s3_key_id = None
            self.invalid = True
            self.display_name = "Unprovisioned S3 Credential"

    @property
    def s3_secret_key(self):
        value = self._s3_secret_key
        if value is None:
            raise AttributeError()
        return value

    @s3_secret_key.setter
    def s3_secret_key(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self._s3_secret_key = value
            if value is None:
                self.invalid = True
        elif value:
            raise ValueError()

    @s3_secret_key.deleter
    def s3_secret_key(self):
        if self.storage_gateway.s3_user_credential_required:
            self._s3_secret_key = None
            self.invalid = True

    def patch(self, **doc):
        doc.pop("display_name", None)
        policies = doc.get("policies", {})

        if self.storage_gateway.s3_user_credential_required:
            if policies is None:
                s3_key_id = None
                s3_secret_key = None
            else:
                s3_key_id = policies.get("s3_key_id", getattr(self, "s3_key_id", None))
                s3_secret_key = policies.get(
                    "s3_secret_key", getattr(self, "s3_secret_key", None)
                )
            if not s3_key_id or not s3_secret_key:
                if not doc.get("invalid", True):
                    raise ValueError("Cannot mark this credential as valid")
                doc["invalid"] = True
            else:
                doc["provisioned"] = True

            doc["display_name"] = self._provisioned_display_name(s3_key_id)
        super().patch(**doc)

    def put(self, **doc):
        invalid = doc.get("invalid")
        policies = doc.get("policies", {})
        s3_key_id = policies.get("s3_key_id")
        s3_secret_key = policies.get("s3_secret_key")

        if self.storage_gateway.s3_user_credential_required:
            if not policies:
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }
            if not (s3_key_id and s3_secret_key):
                if invalid is False:
                    raise ValueError("Cannot mark this credential as valid")
                invalid = True
            else:
                self.provisioned = True
        else:
            if s3_key_id or s3_secret_key:
                raise ValueError("key not allowed for anonymous access")
            policies = {
                "DATA_TYPE": self.POLICY_DATA_TYPE,
            }
            self.provisioned = True
        if invalid is None:
            invalid = False

        doc["invalid"] = invalid
        doc["display_name"] = self._provisioned_display_name(s3_key_id)

        super().put(**doc)

    def _provisioned_display_name(self, s3_key_id):
        if self.storage_gateway.s3_user_credential_required:
            if s3_key_id:
                display_name = s3_key_id
            else:
                display_name = self.unconfigured_display_name
        else:
            display_name = self.anonymous_display_name

        return display_name

    def delete(self):
        self.s3_secret_key = None
        super().delete()

    def __repr__(self):
        return "<S3UserCredential {}>".format(self.id)
