from typing import Dict
import marshmallow as ma

from globus.manager.constants import CollectionType
from .data_type import DataTypeSchema, data_type_schema
from ..patch_schema import create_patch_schema
from ..enum_field import EnumField
from ..cleanup_uuid_mixin import CleanupUUIDMixin
from .path_restrictions import path_restrictions_schema


class CollectionSchema(ma.Schema, CleanupUUIDMixin):
    _DATA_TYPE = "collection#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    id = ma.fields.UUID(dump_only=True)
    domain_name = ma.fields.String()
    connector_id = ma.fields.UUID(dump_only=True)
    manager_url = ma.fields.Url(schemes=["https"], dump_only=True)
    https_url = ma.fields.Url(schemes=["https"], dump_only=True)
    tlsftp_url = ma.fields.Url(schemes=["tlsftp"], dump_only=True)
    collection_type = EnumField(required=True, enum_class=CollectionType)
    display_name = ma.fields.String(
        required=True, validate=[ma.validate.Length(max=128), ma.validate.Regexp(r"\S")]
    )
    identity_id = ma.fields.UUID(required=False)
    storage_gateway_id = ma.fields.UUID()
    mapped_collection_id = ma.fields.UUID(required=False)
    root_path = ma.fields.String(required=False, dump_only=True)
    collection_base_path = ma.fields.String(
        required=True, validate=[ma.validate.Regexp(r"^(/|$HOME)")]
    )
    default_directory = ma.fields.String(validate=[ma.validate.Regexp(r"^(/|$HOME)")])
    public = ma.fields.Boolean(required=True)
    force_encryption = ma.fields.Boolean()
    disable_verify = ma.fields.Boolean()
    user_credential_id = ma.fields.UUID(required=False)
    organization = ma.fields.String(
        validate=[ma.validate.Length(max=1024)], required=False
    )
    department = ma.fields.String(
        validate=[ma.validate.Length(max=1024)], allow_none=True, required=False
    )
    keywords = ma.fields.List(ma.fields.String(), required=False)
    description = ma.fields.String(
        validate=[ma.validate.Length(max=4096)], allow_none=True, required=False
    )
    contact_email = ma.fields.Email(allow_none=True, required=False)
    contact_info = ma.fields.String(
        validate=[ma.validate.Length(max=4096)], allow_none=True
    )
    info_link = ma.fields.Url(allow_none=True, required=False)
    high_assurance = ma.fields.Boolean(dump_only=True)
    authentication_timeout_mins = ma.fields.Int(dump_only=True)
    allow_guest_collections = ma.fields.Boolean(required=False)
    policies = ma.fields.Method(serialize="dump_policies", deserialize="load_policies")
    sharing_restrict_paths = ma.fields.Nested(
        path_restrictions_schema, required=False, allow_none=True
    )
    deleted = ma.fields.Boolean(required=False)

    private = [
        "root_path",
        "collection_base_path",
        "default_directory",
        "sharing_restrict_paths",
        "deleted",
    ]
    uuid_fields = [
        "id",
        "identity_id",
        "storage_gateway_id",
        "user_credential_id",
        "mapped_collection_id",
    ]

    @ma.post_load
    def clean_keywords(self, data, **kwargs):
        res = data.copy()
        keywords = res.get("keywords")
        if keywords:
            res["keywords"] = [kw.strip() for kw in data["keywords"]]
        return res

    @ma.validates("keywords")
    def validate_keywords(self, data, **kwargs):
        if data:
            kw = ",".join(data)
            if len(kw) > 1024:
                raise ma.ValidationError("Longer than maximum length 1024.")

    @ma.validates_schema
    def validate_mapped_and_guest_only_fields(self, data, **kwargs):
        # Patch may not have collection type set, so we can't fully validate
        # without loading the object from the database
        collection_type = data.get("collection_type")
        if collection_type == CollectionType.mapped:
            field = None
            if "user_credential_id" in data:
                field = "user_credential_id"
                raise ma.ValidationError(
                    "Invalid field for mapped collection.", field_name=field
                )
        elif collection_type == CollectionType.guest:
            for field in ["allow_guest_collections", "sharing_restrict_paths"]:
                if field in data and data[field]:
                    raise ma.ValidationError(
                        "Invalid field for guest collection.", field_name=field
                    )

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial)

    def dump_policies(self, obj):
        connector_id = str(ma.utils.get_value(obj, "connector_id"))
        schema = self.connector_policies_schemas.get(connector_id)
        if schema:
            return schema.dump(obj)
        else:
            raise ValueError("XXX: Unable to serialize {} {}".format(obj, connector_id))

    # connectors may register their schema types in this dict, with
    # the connector ID as the key
    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]

    @classmethod
    def register_schema(cls, schema):
        connector_id = schema.connector_id
        cls.connector_policies_schemas[connector_id] = schema


collection_schema = CollectionSchema(load_only=CollectionSchema.private)
collection_private_schema = CollectionSchema()
collection_patch_schema = create_patch_schema(CollectionSchema)

DataTypeSchema.register_schema(collection_schema)
