import enum
import marshmallow as ma
from .data_type import DataTypeSchema

_principal_type = {
    "urn:globus:auth:identity": "identity",
    "urn:globus:groups:id": "group",
}
_principal_re = r"^(urn:globus:auth:identity|urn:globus:groups:id):([a-f0-9-]*)$"


class RoleSchema(ma.Schema):
    """
    The "Role" document type represents the assignment of a role on
    an Endpoint or Collection to a Globus identity or group.
    """

    _DATA_TYPE = "role#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    id = ma.fields.UUID()
    collection = ma.fields.UUID(required=False, allow_none=True)
    principal = ma.fields.String(
        required=True, validate=[ma.validate.Regexp(_principal_re)]
    )
    role = ma.fields.String(
        required=True,
        validate=[
            ma.validate.OneOf(
                [
                    "owner",
                    "administrator",
                    "access_manager",
                    "activity_manager",
                    "activity_monitor",
                    "access_monitor",
                ]
            )
        ],
    )

    @ma.pre_dump
    def adapt_from_model(self, obj, **kwargs):
        """
        Handle divergence between the model, which breaks down some
        document properties into multiple columns to better handle database
        constraints.

        The principal URN is created from the principal_type and identity_id.

        The role prefix is removed since it is clear from presence of the
        non-empty collection property whether it is a collection or endpoint
        role.
        """
        res = {}
        res["id"] = ma.utils.get_value(obj, "id")
        collection = ma.utils.get_value(obj, "resource")
        if collection:
            res["collection"] = collection
        principal_type = ma.utils.get_value(obj, "principal_type")
        if isinstance(principal_type, enum.Enum):
            principal_type = principal_type.value
        res["principal"] = "{}:{}".format(
            [k for k, v in _principal_type.items() if v == principal_type][0],
            ma.utils.get_value(obj, "identity_id"),
        )
        role = ma.utils.get_value(obj, "role")
        if isinstance(role, enum.Enum):
            role = role.name
        res["role"] = role.split("_", 1)[1]

        return res


role_schema = RoleSchema()

DataTypeSchema.register_schema(role_schema)
