from typing import Dict

import marshmallow as ma

from .data_type import DataTypeSchema, data_type_schema
from ..patch_schema import create_patch_schema


class UserCredentialSchema(ma.Schema):
    _DATA_TYPE = "user_credential#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    id = ma.fields.UUID(required=False)
    identity_id = ma.fields.UUID()
    connector_id = ma.fields.UUID()
    storage_gateway_id = ma.fields.UUID()
    username = ma.fields.Str()
    display_name = ma.fields.Str(
        required=False, allow_none=True, validate=[ma.validate.Length(max=256)]
    )
    home_directory = ma.fields.String(dump_only=True)
    invalid = ma.fields.Boolean(required=False)
    provisioned = ma.fields.Boolean(required=False)
    deleted = ma.fields.Boolean(required=False)
    policies = ma.fields.Method(serialize="dump_policies", deserialize="load_policies")

    private = ["home_directory"]

    connector_policies_schemas = {}  # type: Dict[str, ma.Schema]
    connector_policies_private_schemas = {}  # type: Dict[str, ma.Schema]

    @ma.pre_load
    def ignore_loading_home_directory(self, data, **kwargs):
        """
        Remove the home_directory property when loading. We want to serialize
        this when replying to gridftp login but don't want to care about a
        remotely loaded value.
        """
        data.pop("home_directory", None)
        return data

    def load_policies(self, value):
        return data_type_schema.load(value, partial=self.partial, unknown=ma.EXCLUDE)

    def dump_policies(self, obj):
        connector_id = ma.utils.get_value(obj, "connector_id")
        policies = ma.utils.get_value(obj, "policies")

        if not self.load_only:
            schema = self.connector_policies_private_schemas.get(connector_id)
        else:
            schema = self.connector_policies_schemas.get(connector_id)
        if schema:
            if policies:
                return schema.dump(policies)
            else:
                return schema.dump(obj)
        else:
            raise ValueError("Unable to serialize {} {}".format(obj, connector_id))

    @ma.post_dump
    def remove_deleted(self, data, **kwargs):
        if not data.get("deleted"):
            data.pop("deleted", None)
        return data

    @classmethod
    def register_schema(cls, public=None, private=None):
        if public:
            connector_id = public.connector_id
            cls.connector_policies_schemas[connector_id] = public
            if not private:
                cls.connector_policies_private_schemas[connector_id] = public
        if private:
            connector_id = private.connector_id
            cls.connector_policies_private_schemas[connector_id] = private


user_credential_schema = UserCredentialSchema(load_only=UserCredentialSchema.private)
user_credential_private_schema = UserCredentialSchema()
user_credential_patch_schema = create_patch_schema(UserCredentialSchema)
DataTypeSchema.register_schema(UserCredentialSchema)
