from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


BOX_CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"


class BoxAppAuth(ma.Schema):
    publicKeyID = ma.fields.Str(required=True)
    privateKey = ma.fields.Str(required=True)
    passphrase = ma.fields.Str(required=True)


class BoxAppSettingsSchema(ma.Schema):
    clientID = ma.fields.Str(required=True)
    clientSecret = ma.fields.Str(required=True)
    appAuth = ma.fields.Nested(BoxAppAuth, required=True)


class BoxJsonConfigSchema(ma.Schema):
    boxAppSettings = ma.fields.Nested(BoxAppSettingsSchema, required=True)
    enterpriseID = ma.fields.Str(required=True)
    private = ["boxAppSettings", "enterpriseID"]


class BoxStoragePoliciesSchema(BoxJsonConfigSchema):
    _DATA_TYPE = "box_storage_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    connector_id = BOX_CONNECTOR_ID


class BoxCollectionPoliciesSchema(ma.Schema):
    _DATA_TYPE = "box_collection_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


class BoxUserCredentialPoliciesSchema(ma.Schema):
    _DATA_TYPE = "box_user_credential_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    private = []  # type: List[str]
    connector_id = BOX_CONNECTOR_ID


box_storage_policies_private_schema = BoxStoragePoliciesSchema()
box_storage_policies_schema = BoxStoragePoliciesSchema(
    load_only=BoxStoragePoliciesSchema.private
)
box_user_credential_policies_schema = BoxUserCredentialPoliciesSchema(
    load_only=BoxUserCredentialPoliciesSchema.private
)
box_user_credential_policies_private_schema = BoxUserCredentialPoliciesSchema()
box_collection_policies_schema = BoxCollectionPoliciesSchema()

DataTypeSchema.register_schema(box_storage_policies_schema)
DataTypeSchema.register_schema(box_user_credential_policies_schema)
DataTypeSchema.register_schema(box_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=box_storage_policies_schema, private=box_storage_policies_private_schema
)

UserCredentialSchema.register_schema(
    public=box_user_credential_policies_schema,
    private=box_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(box_collection_policies_schema)
