from typing import List

import marshmallow as ma

from ..api.collection import CollectionSchema
from ..api.data_type import DataTypeSchema
from ..api.storage_gateway import StorageGatewaySchema
from ..api.user_credential import UserCredentialSchema


GOOGLE_CLOUD_STORAGE_CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"


class GoogleCloudStorageProjectSchema(ma.Schema):
    projectId = ma.fields.Str()
    name = ma.fields.Str()


class GoogleCloudStoragePoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_cloud_storage_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    client_id = ma.fields.Str(required=True, allow_none=True)
    secret = ma.fields.Str(required=True, allow_none=True)
    buckets = ma.fields.List(ma.fields.String(), required=False, allow_none=True)
    projects = ma.fields.List(ma.fields.String(), required=False, allow_none=True)
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        doc="URL of the auth callback that must be registered on the Google "
        "API console for the application client_id in order to process "
        "Google credentials.",
    )

    private = ["client_id", "secret"]
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID


class GoogleCloudStorageCollectionPoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_cloud_storage_collection_policies#1.0.0"
    DATA_TYPE = ma.fields.String(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    project = ma.fields.String()
    private = []  # type: List[str]
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID


class GoogleCloudStorageUserCredentialPoliciesSchema(ma.Schema):
    _DATA_TYPE = "google_cloud_storage_user_credential_policies#1.0.0"
    DATA_TYPE = ma.fields.Str(
        required=True,
        allow_none=False,
        default=_DATA_TYPE,
        validate=[ma.validate.Equal(_DATA_TYPE)],
    )
    sub = ma.fields.Str()
    email = ma.fields.Email()
    access_token = ma.fields.Str(required=False, allow_none=True)
    refresh_token = ma.fields.Str(required=False, allow_none=True)
    token_expiry = ma.fields.AwareDateTime()
    scopes = ma.fields.List(ma.fields.Str)
    projects = ma.fields.List(ma.fields.Nested(GoogleCloudStorageProjectSchema))

    private = ["access_token", "refresh_token"]
    connector_id = GOOGLE_CLOUD_STORAGE_CONNECTOR_ID


google_cloud_storage_user_credential_policies_private_schema = (
    GoogleCloudStorageUserCredentialPoliciesSchema()
)
google_cloud_storage_user_credential_policies_schema = GoogleCloudStorageUserCredentialPoliciesSchema(
    load_only=GoogleCloudStorageUserCredentialPoliciesSchema.private
)
google_cloud_storage_policies_private_schema = GoogleCloudStoragePoliciesSchema(
    exclude=["auth_callback"]
)
google_cloud_storage_policies_schema = GoogleCloudStoragePoliciesSchema(
    load_only=GoogleCloudStoragePoliciesSchema.private
)
google_cloud_storage_collection_policies_private_schema = (
    GoogleCloudStorageCollectionPoliciesSchema()
)
google_cloud_storage_collection_policies_schema = GoogleCloudStorageCollectionPoliciesSchema(
    load_only=GoogleCloudStorageCollectionPoliciesSchema.private
)

DataTypeSchema.register_schema(google_cloud_storage_policies_schema)
DataTypeSchema.register_schema(google_cloud_storage_user_credential_policies_schema)
DataTypeSchema.register_schema(google_cloud_storage_collection_policies_schema)

StorageGatewaySchema.register_schema(
    public=google_cloud_storage_policies_schema,
    private=google_cloud_storage_policies_private_schema,
)

UserCredentialSchema.register_schema(
    public=google_cloud_storage_user_credential_policies_schema,
    private=google_cloud_storage_user_credential_policies_private_schema,
)

CollectionSchema.register_schema(google_cloud_storage_collection_policies_schema)
