import contextlib
import datetime
from numbers import Number
import os

from flask import current_app, g
import marshmallow as ma

from globus.manager.constants import ConfigurationObjectTag
import globus.manager.api.models as apimodels
from globus.manager.api.schemas.appsync.appsync_data import appsync_data_schema
from globus.manager.utils.interprocess_lock import InterprocessLock

import globus_sdk

from .query import (
    GraphQLArgument,
    GraphQLError,
    GraphQLQuery,
    GraphQLQueryType,
    GraphQLSelection,
    GraphQLVariable,
    GraphQLVariables,
)
from .subscription import Subscription, SubscriptionInfo


class GraphQLResponse(globus_sdk.GlobusHTTPResponse):
    """
    Class used to parse results from the GraphQL service. Note that normally
    the GraphQL API returns errors within a HTTP 200 message response. This
    extracts some bits from the HTTP headers which may indicate some error
    that happens at a higher level than the GraphQL protocol.
    """

    def __init__(self, http_response, *args, **kwargs):
        super().__init__(http_response, *args, **kwargs)
        self.amazon_request_id = http_response.headers.get("x-amzn-RequestId")
        self.amazon_error_type = http_response.headers.get("x-amzn-ErrorType")
        self.amazon_message = self.data.get("message")


class GraphQLApiGatewayError(globus_sdk.GlobusAPIError):
    """
    GCS uses an API Gateway frontend to the GraphQL service. It can return
    some types of errors when parts of the system is down or some serious
    problems with the input (bad auth token, bad URL path).
    """

    def __init__(self, r):
        super().__init__(r)
        self.amazon_request_id = r.headers.get("x-amzn-RequestId")
        self.amazon_error_type = r.headers.get("x-amzn-ErrorType")


class GraphQLClient(globus_sdk.base.BaseClient):
    model_order = [
        ConfigurationObjectTag.KEY_CHAIN,
        ConfigurationObjectTag.ENDPOINT,
        ConfigurationObjectTag.WILDCARD_DOMAIN,
        ConfigurationObjectTag.DOMAIN,
        ConfigurationObjectTag.ACME_ACCOUNT,
        ConfigurationObjectTag.CERTIFICATE,
        ConfigurationObjectTag.STORAGE_GATEWAY,
        ConfigurationObjectTag.USER_CREDENTIAL,
        ConfigurationObjectTag.COLLECTION,
        ConfigurationObjectTag.ROLE,
    ]
    ENDPOINT_SELECTIONS = [
        "owner_id",
        "deleted",
        GraphQLSelection(
            "configuration_objects",
            selections=[
                GraphQLSelection(
                    "items",
                    selections=[
                        "id",
                        "tag",
                        "serial",
                        "created_at",
                        "updated_at",
                        "configuration",
                        "deleted",
                    ],
                ),
                "next_page_token",
            ],
        ),
    ]
    error_class = GraphQLApiGatewayError
    default_response_class = GraphQLResponse

    def __init__(self, owner_id, *args, **kwargs):
        self.owner_id = owner_id
        if "HOME" in current_app.config:
            self.subscription_lock = InterprocessLock(
                "subscription_pause", os.path.join(current_app.config["HOME"], "locks")
            )
        else:
            self.subscription_lock = None

        super().__init__("appsync", *args, **kwargs)

    @staticmethod
    def _awsdatetime(since):
        if isinstance(since, Number):
            since = datetime.datetime.fromtimestamp(since, datetime.timezone.utc)
        if not isinstance(since, datetime.datetime):
            raise TypeError("since must be numeric or datetime")
        offset = since.utcoffset()
        if offset is None:
            offset = datetime.timedelta(seconds=0)
        adjusted = (since + offset).replace(tzinfo=None)
        since_string = adjusted.isoformat() + "Z"
        return since_string

    def query_all(self, item_filter=None, since=None, tag=None, id=None):
        """
        Query AppSync for configuration information. This can be filtered
        by time, object tag, or id. This function process pagination
        internally.

        The since value can be either a numerical unix time offset or a
        datetime.datetime object. If it is a datetime.datetime and does not
        contain a timezone it is interpreted as a UTC time.
        """
        has_next_page = True
        next_page_token = None
        all_items = []
        iterations = 0

        while has_next_page:
            items, next_page_token = self._query(
                item_filter=item_filter,
                since=since,
                tag=tag,
                id=id,
                page_token=next_page_token,
            )
            all_items += items
            has_next_page = next_page_token
            iterations += 1

        current_app.logger.debug(
            "%s pages of data with %s items", iterations, len(all_items)
        )

        return all_items

    def _query(self, item_filter=None, since=None, tag=None, id=None, page_token=None):
        """
        Query AppSync for configuration information. This can be filtered
        by time, object tag, or id, and can be paginated.

        The since value can be either a numerical unix time offset or a
        datetime.datetime object. If it is a datetime.datetime and does not
        contain a timezone it is interpreted as a UTC time.

        If item_filter is set to a callable, it will be called with a list
        of responses to determine which ones to process into models to return
        """
        query_vars = []
        query_args = []

        if since is not None:
            since = self._awsdatetime(since)
            since_var = GraphQLVariable("since", "AWSDateTime", since)
            query_vars.append(since_var)
            query_args.append(GraphQLArgument("since", since_var))
        if tag is not None:
            if not isinstance(tag, ConfigurationObjectTag):
                tag = ConfigurationObjectTag(tag)
            tag_var = GraphQLVariable("tag", "ConfigurationObjectTag", tag)
            query_vars.append(tag_var)
            query_args.append(GraphQLArgument("tag", tag_var))

        if id is not None:
            id_var = GraphQLVariable("id", "ID", str(id))
            query_vars.append(id_var)
            query_args.append(GraphQLArgument("id", id_var))

        if page_token is not None:
            page_token_var = GraphQLVariable("page_token", "String", page_token)
            query_vars.append(page_token_var)
            query_args.append(GraphQLArgument("page_token", page_token_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.QUERY,
            GraphQLSelection(
                "Endpoint",
                selections=[
                    "owner_id",
                    "deleted",
                    GraphQLSelection(
                        "configuration_objects",
                        arguments=query_args,
                        selections=[
                            GraphQLSelection(
                                "items",
                                selections=[
                                    "owner_id",
                                    "id",
                                    "tag",
                                    "serial",
                                    "created_at",
                                    "updated_at",
                                    "configuration",
                                    "deleted",
                                ],
                            ),
                            "next_page_token",
                        ],
                    ),
                ],
            ),
            query_variables,
        )

        res = self.post(
            "/graphql",
            json_body=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "Endpoint")

        if item_filter:
            res["items"] = item_filter(res["items"])

        return (
            res["items"],
            res.get("configuration_objects", {}).get("next_page_token"),
        )

    @staticmethod
    def process_keychain_item(item):
        try:
            current_app.keychain.update(item)
            return current_app.keychain
        except Exception as e:
            current_app.logger.error(
                "Error processing keychain %s: %s", item["id"], str(e)
            )

    @staticmethod
    def _process_item(item, model_class, **kwargs):
        old_model = model_class.query.filter(model_class.id == item["id"]).one_or_none()

        if old_model:
            try:
                if item.get("deleted", False):
                    old_model.deleted = True
                else:
                    old_model.update(item)
                return old_model
            except Exception as e:
                current_app.logger.error(
                    "Error processing update to object %s: %s", item["id"], str(e)
                )
        else:
            try:
                if item.get("deleted", False):
                    current_app.logger.error(
                        "Ignoring previously unknown object %s which is deleted=%s",
                        item["id"],
                        item.get("deleted", False),
                    )
                else:
                    return model_class(appsync_data=item, **kwargs)
            except Exception as e:
                current_app.logger.error(
                    "Error processing previously unknown object %s: %s",
                    item["id"],
                    str(e),
                )

    @staticmethod
    def process_endpoint_item(item):
        return GraphQLClient._process_item(item, apimodels.Endpoint)

    @staticmethod
    def process_wildcard_domain_item(item):
        return GraphQLClient._process_item(item, apimodels.Domain)

    @staticmethod
    def process_acme_account_item(item):
        return GraphQLClient._process_item(item, apimodels.AcmeAccount)

    @staticmethod
    def process_certificate_item(item):
        return GraphQLClient._process_item(item, apimodels.Certificate)

    @staticmethod
    def process_storage_gateway_item(item):
        try:
            configuration = current_app.keychain.decrypt(item["configuration"])
        except Exception:
            current_app.logger.error("Error decrypting configuration %s", item["id"])
            return

        model_cls = apimodels.StorageGateway.polymorphic_class(
            configuration["connector_id"]
        )

        return GraphQLClient._process_item(item, model_cls)

    @staticmethod
    def process_user_credential_item(item):
        kwargs = {}
        try:
            configuration = current_app.keychain.decrypt(item["configuration"])
        except Exception:
            current_app.logger.error("Error decrypting configuration %s", item["id"])
            return

        storage_gateway = apimodels.StorageGateway.query.filter(
            apimodels.StorageGateway.id == configuration.get("storage_gateway_id", "")
        ).one_or_none()
        if not storage_gateway:
            return None
        kwargs["storage_gateway"] = storage_gateway

        model_cls = apimodels.UserCredential.polymorphic_class(
            storage_gateway.connector_id
        )

        return GraphQLClient._process_item(item, model_cls, **kwargs)

    @staticmethod
    def process_collection_item(item):
        kwargs = {}
        try:
            configuration = current_app.keychain.decrypt(item["configuration"])
        except Exception:
            current_app.logger.error("Error decrypting configuration %s", item["id"])
            return

        storage_gateway_id = configuration.get("storage_gateway_id")
        if not storage_gateway_id:
            return

        storage_gateway = apimodels.StorageGateway.query.filter(
            apimodels.StorageGateway.id == storage_gateway_id
        ).one_or_none()
        if not storage_gateway:
            return None
        kwargs["storage_gateway"] = storage_gateway

        mapped_collection_id = configuration.get("mapped_collection_id")
        if mapped_collection_id:
            mapped_collection = apimodels.Collection.query.filter(
                apimodels.Collection.id == mapped_collection_id
            ).one_or_none()
            if not mapped_collection:
                return None
            kwargs["mapped_collection"] = mapped_collection

        user_credential_id = configuration.get("user_credential_id")
        if user_credential_id:
            user_credential = apimodels.UserCredential.query.filter(
                apimodels.UserCredential.id == user_credential_id
            ).one_or_none()
            if not user_credential:
                return None
            kwargs["user_credential"] = user_credential

        endpoint = apimodels.Endpoint.query.one_or_none()
        if not endpoint:
            return None

        model_cls = apimodels.Collection.polymorphic_class(storage_gateway.connector_id)

        return GraphQLClient._process_item(item, model_cls, **kwargs)

    @staticmethod
    def process_role_item(item):
        try:
            configuration = current_app.keychain.decrypt(item["configuration"])
        except Exception:
            current_app.logger.error("Error decrypting configuration %s", item["id"])
            return

        collection_id = configuration.get("resource")
        if collection_id:
            if item.get("deleted", True):
                current_app.logger.debug(
                    "Role %s is deleted, not checkig if we can resolve collection_id %s",
                    item["id"],
                    collection_id,
                )
            else:
                current_app.logger.debug("Checking for collection %s", collection_id)
                collection = apimodels.Collection.query.filter(
                    apimodels.Collection.id == collection_id
                ).one_or_none()
                if not collection:
                    current_app.logger.debug(
                        "Deferring %s %s, collection does %s not resolve",
                        item["tag"],
                        item["id"],
                        collection_id,
                    )
                    return None

        return GraphQLClient._process_item(item, apimodels.Role)

    @staticmethod
    def process_item(item):
        dispatch = {
            ConfigurationObjectTag.ACME_ACCOUNT: GraphQLClient.process_acme_account_item,
            ConfigurationObjectTag.CERTIFICATE: GraphQLClient.process_certificate_item,
            ConfigurationObjectTag.COLLECTION: GraphQLClient.process_collection_item,
            ConfigurationObjectTag.ENDPOINT: GraphQLClient.process_endpoint_item,
            ConfigurationObjectTag.KEY_CHAIN: GraphQLClient.process_keychain_item,
            ConfigurationObjectTag.ROLE: GraphQLClient.process_role_item,
            ConfigurationObjectTag.STORAGE_GATEWAY: GraphQLClient.process_storage_gateway_item,
            ConfigurationObjectTag.USER_CREDENTIAL: GraphQLClient.process_user_credential_item,
            ConfigurationObjectTag.WILDCARD_DOMAIN: GraphQLClient.process_wildcard_domain_item,
        }
        return dispatch[item["tag"]](item)

    def query(self, item_filter=None, since=None, tag=None, id=None, page_token=None):
        """
        Query AppSync for configuration information. This can be filtered
        by time, object tag, or id, and can be paginated.

        The since value can be either a numerical unix time offset or a
        datetime.datetime object. If it is a datetime.datetime and does not
        contain a timezone it is interpreted as a UTC time.

        If item_filter is set to a callable, it will be called with a list
        of responses to determine which ones to process into models to return
        """
        return self._query(
            item_filter=item_filter, since=since, tag=tag, id=id, page_token=page_token
        )

    def create_object(self, model):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        id_var = GraphQLVariable("id", "ID!", model.id)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        tag_var = GraphQLVariable("tag", "ConfigurationObjectTag!", model.tag)
        query_vars.append(tag_var)
        query_args.append(GraphQLArgument("tag", tag_var))

        configuration_var = GraphQLVariable(
            "configuration", "String!", model.configuration
        )
        query_vars.append(configuration_var)
        query_args.append(GraphQLArgument("configuration", configuration_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectCreate",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "/graphql",
            json_body=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "ObjectCreate")

        assert len(res["items"]) == 1
        for o in res["items"]:
            o["tag"] = ConfigurationObjectTag(o["tag"])
            model.update(o)

        return model

    def update_object(self, model, ignore_response=False):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        id_var = GraphQLVariable("id", "ID!", model.id)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        serial_var = GraphQLVariable("serial", "Int!", model.serial + 1)
        query_vars.append(serial_var)
        query_args.append(GraphQLArgument("serial", serial_var))

        configuration_var = GraphQLVariable(
            "configuration", "String!", model.configuration
        )
        query_vars.append(configuration_var)
        query_args.append(GraphQLArgument("configuration", configuration_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectUpdate",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "/graphql",
            json_body=dict(query=str(query), variables=query_variables.variables()),
        )

        if not ignore_response:
            res = self._parse_result(res, "ObjectUpdate")

            assert len(res["items"]) == 1
            for o in res["items"]:
                o["tag"] = ConfigurationObjectTag(o["tag"])
                model.update(o)

        return model

    def delete_object(self, model=None, id=None):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        if model:
            id_val = model.id
        elif id is not None:
            id_val = id
        else:
            raise RuntimeError()

        id_var = GraphQLVariable("id", "ID!", id_val)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectDelete",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "/graphql",
            json_body=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "ObjectDelete")
        return res

    def subscribe(self, **kwargs):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.SUBSCRIPTION,
            GraphQLSelection(
                "Endpoint", arguments=query_args, selections=self.ENDPOINT_SELECTIONS
            ),
            query_variables,
        )

        try:
            res = self.post(
                "/graphql",
                json_body=dict(query=str(query), variables=query_variables.variables()),
            )
        except GraphQLApiGatewayError as e:
            # This is a API Gateway Level Error, it probably has an access token
            # in the message body, so we'll just deal with the headers
            request_id = getattr(g, "request_id", None) if g else None

            current_app.logger.info(
                "graphql_exception: gcs_request_id=%s "
                "graphql_request_id=%s error=%s",
                request_id,
                e.amazon_request_id,
                e.amazon_error_type,
            )
            raise

        if res.get("extensions"):
            subscription_info = SubscriptionInfo(res.get("extensions")["subscription"])
            return Subscription(subscription_info, **kwargs)
        elif res.get("errors"):
            exc = GraphQLError(res["errors"], None)
            request_id = getattr(g, "request_id", None) if g else None

            current_app.logger.info(
                "graphql_exception: gcs_request_id=%s "
                'graphql_request_id=%s error=%s message="%s"',
                request_id,
                res.amazon_request_id,
                exc.errors[0]["errorType"],
                exc.message,
            )
            raise exc

    @contextlib.contextmanager
    def pause_notifications(self):
        """
        Pause subscription message processing
        """
        with self.subscription_lock.shared():
            yield

    def _parse_result(self, res, wrapper_property, unknown=ma.RAISE):
        data = res.get("data", {}).get(wrapper_property, {})
        if data:
            try:
                data = appsync_data_schema.load(data, unknown=unknown)
            except Exception as e:
                current_app.logger.error(
                    "Error parsing result from appsync: %s", str(e)
                )

        if res.get("errors"):
            exc = GraphQLError(res["errors"], data)
            request_id = getattr(g, "request_id", None) if g else None

            current_app.logger.info(
                "graphql_exception: gcs_request_id=%s "
                'graphql_request_id=%s error=%s message="%s"',
                request_id,
                res.amazon_request_id,
                exc.errors[0]["errorType"],
                exc.message,
            )
            raise exc

        return data

    @staticmethod
    def item_sort_key(item):
        sort_key = GraphQLClient.model_order.index(item["tag"])
        if item["tag"] == ConfigurationObjectTag.COLLECTION:
            configuration = current_app.keychain.decrypt(item["configuration"])
            if configuration["collection_type"] == "guest":
                sort_key += 0.5
        return sort_key

    @staticmethod
    def commit_items_to_session(session, items):
        """
        After querying appsync, construct (or delete) the model objects based
        on the configuration_object items returned from appsync.

        :param session:
            The session to which this adds or deletes the models. The changes
            are committed in this function.

        :param items:
            A list of items from appsync.

        :return:
            Returns a list of items which could not be safely added to the
            database because of a dangling foreign key constraint. They can be
            resubmitted this function along with later notification data.
        """
        old_autoflush = session.autoflush
        deferred_items = []
        session.autoflush = False

        previous_key = -1
        for m in sorted(items, key=GraphQLClient.item_sort_key):
            this_key = GraphQLClient.item_sort_key(m)
            if this_key != previous_key and session.dirty or session.new:
                session.commit()
            model = GraphQLClient.process_item(m)
            if model:
                if isinstance(model, apimodels.db.Model):
                    if m.get("deleted"):
                        session.delete(model)
                    else:
                        session.add(model)
            else:
                deferred_items.append(m)
            previous_key = this_key

        session.autoflush = old_autoflush
        if session.new or session.dirty or session.deleted:
            session.commit()

        return deferred_items
