import json
import logging
import os
import sched
import threading

import click

from globus.manager.version import __version__
from .certificate_manager import CertificateManager
from .gridftp_key_manager import GridFTPKeyManager
from .notification_manager import NotificationManager

from globus.manager.api.app import make_app
import globus.manager.api.models as m
from globus.manager.utils.keychain import NodeKey
from globus.manager.utils.interprocess_lock import InterprocessLock
from globus.manager.utils.platform_info import PlatformInfo


@click.command(
    name="assistant",
    short_help="Background processing for GCS Manager",
    hidden=True,
    help="Background processing for GCS Manager",
)
@click.version_option(version=__version__, prog_name="globus-connect-server")
@click.help_option("--help", "-h")
@click.option("--debug", "-d", is_flag=True)
def assistant_command(debug):
    assistant = Assistant(debug)
    assistant.run()


class Assistant(sched.scheduler):
    def __init__(self, debug):
        self.__interrupt_delay_event = threading.Event()
        super().__init__(delayfunc=self.interruptable_delay)

        self.app = make_app(cli=True, app_config="globus.manager.api.config.assistant")
        m.db.init_app(self.app)

        with self.app.app_context():
            if debug:
                self.app.logger.setLevel(logging.DEBUG)
            else:
                level = os.environ.get("GCS_MANAGER_LOG_LEVEL", "INFO")
                self.app.logger.setLevel(level)

            for c in m.Configuration.query.all():
                if c.name == "node_key":
                    self.app.config["NODE_KEY"] = NodeKey(**json.loads(c.value))
                else:
                    self.app.config[c.name.upper()] = c.value

            items = self.app.appsync_client.query_all(id="KEY_CHAIN")
            for item in items:
                self.app.appsync_client.process_item(item)

            lock = InterprocessLock(
                "assistant", lockdir=os.path.join(self.app.config["HOME"], "locks")
            )

            with lock.exclusive():
                host_endpoint_id = (
                    m.Configuration.query.filter_by(name="host_endpoint_id").one().value
                )
                if host_endpoint_id is not None:
                    fingerprint = PlatformInfo().fingerprint()
                    ep = m.Endpoint.query.one()
                    try:
                        old_fp = ep.fingerprint
                    except AttributeError:
                        old_fp = None

                    if old_fp != fingerprint:
                        ep.fingerprint = fingerprint

                        try:
                            self.app.appsync_client.update_object(ep)
                            m.db.session.add(ep)
                            m.db.session.commit()
                        except Exception:
                            self.app.logger.warn("Unable to update fingerprint")

                    self.certificate_manager = CertificateManager(self, self.app)
                    self.gridftp_key_manager = GridFTPKeyManager(self, self.app)
                    self.notification_manager = NotificationManager(self, self.app)

    def enterabs(self, *args, **kwargs):
        """
        Replacement for enterabs that causes the delay function to be woken
        up when an event is added to the queue. This may cause spurious wakeups
        if the new event is far in the future, but that should not be a serious
        issue for this application.
        """
        res = super().enterabs(*args, **kwargs)
        self.__interrupt_delay_event.set()
        return res

    def interruptable_delay(self, secs):
        """
        Replacemenet for time.sleep as a delay function for sched.scheduler.

        Normally, in the scheduler.run loop, it grabs the time delay for the
        lead event and then sleeps, and then pops the lead event off to execute
        it. However, if during the sleep, another event with a shorter delay
        time is added to the head of the queue, that item must wait until the
        older event's sleep has completed.

        This version will use a threading event to wait for the delay time. If
        the event is signalled, then this function returns immediately. Since
        the run function checks the queue for the lead event's timeout before
        processing, it should be able to recover and continue processing the
        shorter delay time.
        """
        if self.__interrupt_delay_event.wait(secs):
            # early wakeup due to event
            self.__interrupt_delay_event.clear()


if __name__ == "__main__":
    assistant_command()
