from contextlib import suppress
import datetime

import marshmallow as ma

import globus.manager.api.models as models
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.api.schemas.appsync.appsync_data import appsync_data_schema


class NotificationManager:
    """
    This class handles maintains synchronization between appsync and the
    local sqlite database.
    """

    PAGINATION_DELAY = 5
    """
    Delay (in seconds) to wait after a query returns a paginated response to
    perform another query.
    """

    ERROR_DELAY = 30
    """
    Delay (in seconds) to wait after a query or subscription call fails before
    trying again.
    """

    def __init__(self, scheduler, app):
        """
        Initialize a NotificationManager object to handle an appsync subscription.

        :param sched.scheduler scheduler:
            A scheduler which this class will use to register timed events.
        :param GCSManagerApp app:
            An app which this class will use to interact with appsync, the
            database, locks and logs.
        """
        self.scheduler = scheduler
        "Scheduler to register events with"

        self.app = app
        "Flask application this object is a part of"

        self.deferred_items = []
        """
        List of appsync items which can not yet be processed because they
        have a foreign key reference to some other object which we don't
        (yet) have in the database. Ordering of notifications isn't well
        defined in appsync
        """

        self.last_item_timestamp = None
        """
        Highest value of the updated_at property for an object we've
        received in response to an appsync query or notification. This is
        used to make resubscriptions only query for recent changes.
        """
        self.appsync_client = app.appsync_client

        self.lock = self.appsync_client.subscription_lock
        """
        Lock to use when processing query results. This locks out the GCS
        manager process from processing requests that would interact with
        the database while we're processing changes that might make many
        changes to the database.
        """

        self.locked = False
        """
        Flag to indicate whether the lock is currently held
        """

        self.page_token_timestamp = None
        """
        Value of self.last_item_timestamp when a pagination query
        began. This value cannot change while pagination is in progress.
        """

        self.page_token = None
        """
        Query response pagination token.
        """

        self.notifier = None
        """
        Value returned from the appsync client's subscribe method.
        """
        self.scheduler.enter(0, 0, self.query_and_subscribe)

    def query_and_subscribe(self):
        """
        Event to resynchronize state in the local database with the appsync
        database. It performs a full query of the objects in appsync, and
        then subscribes for changes to the appsync endpoint.

        This is triggered when the NotificationManager is first started, and
        then also when the appsync subscription is disconnected (which happens
        approximately every 24 hours).

        To be easy on our appsync implementation, when a paginated result
        is returned from the appsync query, the next page request is slightly
        delayed to a further event.

        If there's an error querying or subscribing, a longer delay is used.
        If this function returns successfully, then a new event is registered,
        either another call to this event, or via the expected eventual
        on_disconnect call.
        """
        with self.app.app_context(), self.lock.exclusive():
            self.app.logger.debug("query_and_subscribe enter")
            old_subscription = None
            try:
                endpoint = models.Endpoint.query.one_or_none()
                old_subscription = endpoint.subscription_id
            except Exception as e:
                self.app.logger.warning(
                    "Error checking endpoint subscription_id: %s", str(e)
                )
                pass

            has_next_page = True

            since = None
            if self.page_token:
                since = self.page_token_timestamp
            elif self.last_item_timestamp:
                since = self.last_item_timestamp - datetime.timedelta(minutes=5)
                self.page_token_timestamp = since
            while has_next_page:
                try:
                    items, self.page_token = self.appsync_client.query(
                        since=since, page_token=self.page_token
                    )
                    self.process_items(items, old_subscription)
                    has_next_page = bool(self.page_token)
                except Exception as e:
                    self.app.logger.warning("Error syncing state: %s", str(e))
                    self.scheduler.enter(self.ERROR_DELAY, 0, self.query_and_subscribe)
                    return
                if self.page_token:
                    # If we're hitting pagination issues, let's take it easy
                    # on appsync
                    self.scheduler.enter(
                        self.PAGINATION_DELAY, 0, self.query_and_subscribe
                    )
                    return

            try:
                if not self.notifier:
                    self.notifier = self.appsync_client.subscribe(
                        on_message=self.on_message, on_disconnect=self.on_disconnect
                    )
            except Exception as e:
                self.app.logger.error("Error processing subscription %s", str(e))

            if not self.notifier:
                self.app.logger.warning("Unable to subscribe for notifications")
                self.scheduler.enter(self.ERROR_DELAY, 0, self.query_and_subscribe)

    def process_items(self, items, old_subscription):
        """
        Add items from the appsync response or change notification to the local
        database, updating local configuration files if necessary.

        If there are any deferred changes from a previous query or notifications,
        they may be processed if their foreign keys are now valid. Any objects
        which can not be processed with the current database state will be
        stashed in self.deferred_items for later processing.

        :param items: List of items returned from appsync
        :param old_subscription:
            Subscription ID of the endpoint before the query result or
            notification data has been processed. This is used to update
            configurations that depend on whether the endpoint has a
            subscription or not.
        """
        items.extend(self.deferred_items)
        new_subscription = None

        changed_collection_ids = [
            i["id"] for i in items if i["tag"] == ConfigurationObjectTag.COLLECTION
        ]
        changed_tags = set(i["tag"] for i in items)

        changed_endpoint = bool(ConfigurationObjectTag.ENDPOINT in changed_tags)
        changed_certificate = bool(ConfigurationObjectTag.CERTIFICATE in changed_tags)

        if items:
            newest_change = sorted([i["updated_at"] for i in items], reverse=True)[0]
            if (
                self.last_item_timestamp is None
                or newest_change > self.last_item_timestamp
            ):
                self.last_item_timestamp = newest_change

        # Update models in local database
        with self.appsync_client.subscription_lock.exclusive():
            self.deferred_items = self.appsync_client.commit_items_to_session(
                models.db.session, items
            )

        if changed_endpoint:
            # If endpoint changed, a subscription state change requires us
            # to replace all collection vhosts
            try:
                endpoint = models.Endpoint.query.one_or_none()
                new_subscription = getattr(endpoint, "subscription_id", None)
                self.app.subscription = new_subscription
            except Exception:
                pass
        else:
            new_subscription = old_subscription

        if changed_certificate:
            # If cert changed, need to update the local config
            try:
                certificate = models.Certificate.query.one_or_none()
                certificate.write_config(certificate.deleted)
            except Exception:
                self.app.logger.warning("Unable to update %s", certificate)

        collection_query = models.Collection.query
        if bool(old_subscription) == bool(new_subscription):
            # If managed state has not changed, only need to update
            # vhosts for changed collections
            collection_query = collection_query.filter(
                models.Collection.id.in_(changed_collection_ids)
            )

        for collection in collection_query:
            try:
                collection.write_config(collection.deleted)
            except Exception:
                self.app.logger.warning("Unable to update %s vhost config", collection)

    def on_message(self, obj, data):
        """
        Appsync change notification message handler.

        This function is called by the mqtt processor to indicate a change
        notification message has been received. It deserializes ``data`` and
        passes it to ``NotificationManager.process_items`` to process the data
        or schedule it for deferred processing.

        :param obj: This object
        :param data: The notification message from appsync.

        NOTE: Because this is called from the appsync subscription mqtt
        processor's thread, calls to the appsync API can not be done here or
        the process will deadlock.
        """
        with self.lock.exclusive():
            try:
                endpoint = models.Endpoint.query.one_or_none()
                old_subscription = endpoint.subscription_id
            except Exception as e:
                self.app.logger.warning("Exception reading endpoint %s", str(e))
                old_subscription = None

            result = {}
            with suppress(KeyError):
                result = data["data"]["Endpoint"]

            if result:
                parsed = appsync_data_schema.load(result, unknown=ma.EXCLUDE)
                self.app.logger.debug(
                    "Updates from appsync for %s",
                    [i["tag"].value + i["id"] for i in parsed["items"]],
                )

                self.process_items(parsed["items"], old_subscription)
            else:
                self.app.logger.warning("Couldn't parse data %s", data)

    def on_disconnect(self, *args, **kwargs):
        """
        Appsync subscription disconnect event handler.

        This function is called when appsync stops sending responses to the
        ping requests for this objects notifier object. This usually happens
        every 24 hours when the appsync subscription key expires. When this
        occurs, we will synchronize and resubscribe.

        NOTE: Because this is called from the appsync subscription mqtt
        processor's thread, calls to the appsync API can not be done here or
        the process will deadlock.
        """
        # Can't call self.notifier.disconnect() in this thread or paho
        # deadlocks
        self.scheduler.enter(0, 0, self.notifier_disconnect)

    def notifier_disconnect(self):
        """
        Disconnect from the appsync subscription. This is triggered by
        the ``NotificationManager.on_disconnect`` method but is run in a
        different thread so we can close the connection without a deadlock.

        When disconnected, schedule the
        ``NotificationManager.query_and_subscribe`` method to run to update
        state and resume the subscription.
        """
        with self.app.app_context():
            self.notifier.disconnect()
            self.notifier = None
            self.scheduler.enter(0, 0, self.query_and_subscribe)
