import datetime
import json
import os

import click
from flask import current_app, g

import globus.manager.api.models as models

from globus.manager.utils.keychain import NodeKey
from globus.manager.appsync.query import GraphQLError
from globus.manager.api.app import make_app


@click.command(
    "reap", help="Remove data related to deleted database objects", hidden=True
)
@click.option(
    "--age",
    help="Minimum age of deleted objects to remove (in days)",
    type=int,
    metavar="DAYS",
    show_default=True,
    default=2,
)
def reap(age):
    app = make_app(cli=True, app_config="globus.manager.api.config.assistant")
    context = app.app_context()

    with context:
        models.db.init_app(app)
        for c in models.Configuration.query.all():
            if c.name == "node_key":
                app.config["NODE_KEY"] = NodeKey(**json.loads(c.value))
            else:
                app.config[c.name.upper()] = c.value
        items = app.appsync_client.query_all(id="KEY_CHAIN")
        for item in items:
            app.appsync_client.process_item(item)

    age_delta = datetime.timedelta(days=age)
    reaper = Reaper(context)
    reaper.reap(age=age_delta)


class Reaper:
    def __init__(self, context):
        self.context = context

    def reap(self, age: datetime.timedelta = None):
        with self.context:
            g.request_id = "reaper"

            if age is None:
                age = datetime.timedelta(days=2)
            newest = datetime.datetime.now(tz=datetime.timezone.utc) - age
            model_queries = [
                models.Role.query,
                models.Collection.query.filter(models.Collection.is_guest_collection),
                models.UserCredential.query.filter(
                    models.UserCredential.id.notin_(
                        models.db.session.query(
                            models.Collection.user_credential_id
                        ).filter(models.Collection.is_guest_collection)
                    )
                ),
                models.Collection.query.filter(
                    models.Collection.is_mapped_collection,
                    models.Collection.id.notin_(
                        models.db.session.query(
                            models.Collection.mapped_collection_id
                        ).filter(models.Collection.is_guest_collection)
                    ),
                ),
                models.StorageGateway.query.filter(
                    models.StorageGateway.id.notin_(
                        models.db.session.query(models.Collection.storage_gateway_id)
                    )
                ),
            ]

            for mq in model_queries:
                mq = mq.filter(
                    models.ConfigurationObject.deleted,
                    models.ConfigurationObject.updated_at < newest,
                )
                qr = mq.all()

                if qr:
                    tag = qr[0].tag
                    try:
                        appsync_ids = set(
                            i["id"]
                            for i in current_app.appsync_client.query_all(tag=tag)
                        )
                    except Exception as e:
                        current_app.logger.warning("Error querying for data: %s", e)
                        continue

                for model in qr:
                    current_app.logger.debug("Reapable object %s %s", model.id, model)
                    try:
                        if model.id in appsync_ids:
                            current_app.logger.debug(
                                "%s not yet deleted in appsync", model
                            )
                            current_app.appsync_client.delete_object(model)

                        if isinstance(model, models.Collection):
                            model.cleanup_vhost()
                        models.db.session.delete(model)
                        models.db.session.commit()
                    except GraphQLError as e:
                        current_app.logger.warning(
                            "Error reaping object %s %s data: %s", model, str(e), e.data
                        )
                    except Exception as e:
                        current_app.logger.warning(
                            "Error reaping object %s %s", model, str(e)
                        )

            # Check for any vhost files in our httpd/conf.d that are
            # named after collections we don't have records for. If so,
            # remove them.
            stale_vhosts = []
            vhost_dir = os.path.expanduser("~gcsweb/etc/httpd/conf.d")

            for vhost in os.listdir(vhost_dir):
                vhost_id = vhost.split(".conf", 1)[0]
                collection = models.Collection.query.filter(
                    models.Collection.id == vhost_id,
                    ~models.Collection.deleted,  # type: ignore
                ).one_or_none()
                if not collection:
                    stale_vhosts.append(vhost)

            for stale in stale_vhosts:
                current_app.logger.info("Deleting stale vhost file %s", stale)
                os.remove(os.path.join(vhost_dir, stale))
