import ctypes
import ctypes.util
import os
import distro

from globus.manager.version import __version__ as version


class PlatformInfo:
    os_version = None
    openssl_version = None
    gcs_version = None

    def __init__(self):
        name = distro.linux_distribution()[0].lower()
        self.platform = None
        platforms = {
            "redhat": ["centos", "redhat", "red hat", "scientific", "fedora"],
            "debian": ["debian", "ubuntu"],
            "suse": ["suse"],
        }
        forced = os.environ.get("GCS_FORCE_PLATFORM", "").lower()
        if forced in platforms.keys():
            self.platform = forced
        else:
            for p in platforms:
                if [x for x in platforms[p] if name.startswith(x)]:
                    self.platform = p
                    break

    def fingerprint(self):
        fingerprint = {
            "OS": PlatformInfo.__os_version(),
            "GCS": PlatformInfo.__gcs_version(),
        }
        openssl_version = PlatformInfo.__openssl_version()
        if openssl_version:
            fingerprint["OPENSSL"] = openssl_version
        return fingerprint

    @staticmethod
    def __os_version():
        if PlatformInfo.os_version is None:
            uname = os.uname()
            dist = distro.linux_distribution(full_distribution_name=False)
            PlatformInfo.os_version = {
                "name": uname[0],
                "kernel": uname[2],
                "distribution": dist[0].lower(),
                "version": dist[1],
                "architecture": uname[4],
            }

        return PlatformInfo.os_version

    @staticmethod
    def __openssl_version():
        if PlatformInfo.openssl_version is None:
            try:
                openssl_version_num = None
                libcrypto = ctypes.CDLL(ctypes.util.find_library("crypto"))
                if getattr(libcrypto, "SSLeay", None):
                    libcrypto.SSLeay.restype = ctypes.c_long
                    openssl_version_num = libcrypto.SSLeay()
                elif getattr(libcrypto, "OpenSSL_version_num", None):
                    libcrypto.OpenSSL_version_num.restype = ctypes.c_long
                    openssl_version_num = libcrypto.OpenSSL_version_num()
                if openssl_version_num is not None:
                    openssl_patch = (openssl_version_num >> 4) & 0xFF
                    openssl_status = openssl_version_num & 0xF
                    if openssl_status == 0:
                        openssl_status = "dev"
                    elif openssl_status < 0xF:
                        openssl_status = "beta-{}".format(openssl_status)
                    else:
                        openssl_status = "release"
                    PlatformInfo.openssl_version = {
                        "major": (openssl_version_num >> 28) & 0xF,
                        "minor": (openssl_version_num >> 20) & 0xFF,
                        "fix": (openssl_version_num >> 12) & 0xFF,
                        "patch": chr(ord("a") - 1 + openssl_patch),
                        "status": openssl_status,
                    }
            except Exception:
                pass
        return PlatformInfo.openssl_version

    @staticmethod
    def __gcs_version():
        if PlatformInfo.gcs_version is None:
            major, minor, patch = version.split(".", 2)
            label = None

            if "-" in patch:
                patch, label = patch.split("-", 1)

            PlatformInfo.gcs_version = {
                "major": int(major),
                "minor": int(minor),
                "patch": int(patch),
            }
            if label:
                PlatformInfo.gcs_version["status"] = label

        return PlatformInfo.gcs_version


class UnsupportedPlatformException(Exception):
    def __init__(self, name=None):
        if name is None:
            name = PlatformInfo().platform
        super(UnsupportedPlatformException, self).__init__(
            "Unsupported platform {}".format(name)
        )
