"""Add mapped_collection_id column

Revision ID: 93e24e66ac6e
Revises: f9ee46456dfe
Create Date: 2020-02-24 12:58:11.092959

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "93e24e66ac6e"
down_revision = "f9ee46456dfe"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("PRAGMA foreign_keys = OFF")
    op.execute("PRAGMA legacy_alter_table=ON")
    conn = op.get_bind()

    # Add NOT NULL constraint to configuration_object.discriminator
    op.rename_table("configuration_object", "configuration_object_old")
    op.create_table(
        "configuration_object",
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("discriminator", sa.TEXT(), nullable=False),
        sa.Column(
            "tag",
            sa.Enum(
                "ACME_ACCOUNT",
                "CERTIFICATE",
                "COLLECTION",
                "DOMAIN",
                "ENDPOINT",
                "KEY_CHAIN",
                "ROLE",
                "STORAGE_GATEWAY",
                "USER_CREDENTIAL",
                "WILDCARD_DOMAIN",
                name="configurationobjecttag",
            ),
            nullable=False,
        ),
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("serial", sa.INTEGER(), nullable=False),
        sa.Column("configuration", sa.TEXT(), nullable=False),
        sa.Column("deleted", sa.BOOLEAN(), server_default="0", nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute(
        """
        INSERT INTO configuration_object
        (id, created_at, updated_at, discriminator, tag,
         serial, configuration, deleted)
        SELECT id, created_at, updated_at, discriminator, tag,
         serial, configuration,
         case when deleted is null then 0 else 1 end
         from configuration_object_old
        """
    )
    op.drop_table("configuration_object_old")

    # Add missing role constraints
    op.rename_table("role", "role_old")
    op.create_table("role",
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('principal_type', sa.Enum('identity', 'group', name='principaltype'), nullable=False),
        sa.Column('identity_id', sa.TEXT(), nullable=False),
        sa.Column('role', sa.Enum('endpoint_owner', 'endpoint_administrator', 'endpoint_activity_manager', 'endpoint_activity_monitor', 'collection_administrator', 'collection_activity_manager', 'collection_activity_monitor', 'collection_access_manager', name='roletype'), nullable=False),
        sa.Column('resource', sa.TEXT(), nullable=True),
        sa.Column('primary_role', sa.BOOLEAN(), server_default='0', nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['id'], ['configuration_object.id']),
        sa.ForeignKeyConstraint(['resource'], ['collection.id']),
        sa.UniqueConstraint('principal_type', 'identity_id', 'role', 'resource', name='unique_collection_role')
    )

    op.execute(
        """
            INSERT INTO role(id, principal_type, identity_id, role, resource, primary_role)
            SELECT id, principal_type, identity_id, role, resource, primary_role from role_old
        """)
    op.drop_table("role_old")

    # Add mapped_collection_id to collection
    op.drop_table("collection")
    conn.execute(
        "DELETE from configuration_object where tag = :tag",
        tag="COLLECTION")
    op.create_table(
        'collection',
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('domain', sa.TEXT(), nullable=False, unique=True),
        sa.Column('user_credential_id', sa.TEXT(), nullable=True),
        sa.Column('storage_gateway_id', sa.TEXT(), nullable=False),
        sa.Column(
            'allow_guest_collections',
             sa.BOOLEAN(),
             server_default='0',
             nullable=False),
        sa.Column(
            'collection_type',
            sa.Enum('mapped', 'guest',
                name='collectiontype'),
            nullable=False),
        sa.Column("mapped_collection_id", sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(
            ['id'],
            ['configuration_object.id'],
            name='col_id'),
        sa.ForeignKeyConstraint(
            ['mapped_collection_id'],
            ['collection.id'],
            name='col_mapped_collection_id'),
        sa.ForeignKeyConstraint(
            ['storage_gateway_id'],
            ['storage_gateway.id'], ),
        sa.ForeignKeyConstraint(
            ['user_credential_id'],
            ['user_credential.id'], ),
        sa.PrimaryKeyConstraint('id')
    )

    # Remove display_name from user_credential table
    op.rename_table("user_credential", "user_credential_old")
    op.create_table(
        "user_credential",
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("identity_id", sa.TEXT(), nullable=False),
        sa.Column("username", sa.TEXT(), nullable=False),
        sa.Column("invalid", sa.Boolean(), server_default="0", nullable=False),
        sa.Column("storage_gateway_id", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["configuration_object.id"],),
        sa.ForeignKeyConstraint(["storage_gateway_id"], ["storage_gateway.id"],),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("storage_gateway_id", "username", "identity_id"),
    )
    op.execute(
        """
	INSERT INTO user_credential (
	    id, identity_id,
	    username,
	    invalid, storage_gateway_id)
	SELECT id, identity_id,
            username,
	    invalid, storage_gateway_id
	FROM user_credential_old
	"""
    )
    op.drop_table("user_credential_old")
    op.create_index(
        op.f("ix_user_credential_identity_id"),
        "user_credential",
        ["identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_credential_invalid"), "user_credential", ["invalid"], unique=False
    )
    op.create_index(
        op.f("ix_user_credential_username"),
        "user_credential",
        ["username"],
        unique=False,
    )


def downgrade():
    op.execute("PRAGMA foreign_keys = OFF")
    op.execute("PRAGMA legacy_alter_table=ON")
    conn = op.get_bind()

    op.drop_index(op.f("ix_user_credential_username"))
    op.drop_index(op.f("ix_user_credential_invalid"))
    op.drop_index(op.f("ix_user_credential_identity_id"))

    # Restore display_name from user_credential table
    op.add_column(
        "user_credential", sa.Column("display_name", sa.TEXT(), nullable=True)
    )
    # Trust we can recover from appsync
    op.execute("DELETE FROM user_credential")
    conn.execute("DELETE FROM configuration_object WHERE tag = :tag",
        tag="USER_CREDENTIAL")

    # Remove mapped_collection_id from collection
    op.drop_table('collection')
    conn.execute("DELETE FROM configuration_object WHERE tag = :tag",
        tag="COLLECTION")
    op.create_table(
        'collection',
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('domain', sa.TEXT(), nullable=False, unique=True),
        sa.Column('user_credential_id', sa.TEXT(), nullable=True),
        sa.Column('storage_gateway_id', sa.TEXT(), nullable=False),
        sa.Column(
            'allow_guest_collections',
             sa.BOOLEAN(),
             server_default='0',
             nullable=False),
        sa.Column(
            'collection_type',
            sa.Enum('mapped', 'guest',
                name='collectiontype'),
            nullable=False),
        sa.ForeignKeyConstraint(
            ['id'],
            ['configuration_object.id'],
            name='col_id'),
        sa.ForeignKeyConstraint(
            ['storage_gateway_id'],
            ['storage_gateway.id'], ),
        sa.ForeignKeyConstraint(
            ['user_credential_id'],
            ['user_credential.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.rename_table("role", "role_old")
    # Drop role foreign key constraint (lost in a previous migration)
    op.create_table("role",
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('principal_type', sa.Enum('identity', 'group', name='principaltype'), nullable=False),
        sa.Column('identity_id', sa.TEXT(), nullable=False),
        sa.Column('role', sa.Enum('endpoint_owner', 'endpoint_administrator', 'endpoint_activity_manager', 'endpoint_activity_monitor', 'collection_administrator', 'collection_activity_manager', 'collection_activity_monitor', 'collection_access_manager', name='roletype'), nullable=False),
        sa.Column('resource', sa.TEXT(), nullable=True),
        sa.Column('primary_role', sa.BOOLEAN(), server_default='0', nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['resource'], ['collection.id']),
        )

    op.execute(
        """
            INSERT INTO role(id, principal_type, identity_id, role, resource, primary_role)
            SELECT id, principal_type, identity_id, role, resource, primary_role from role_old
        """)
    op.drop_table("role_old")

    # Remove NOT NULL constraint from configuration_object.discriminator
    op.rename_table("configuration_object", "configuration_object_old")
    op.create_table(
        "configuration_object",
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.Column("discriminator", sa.TEXT(), nullable=True),
        sa.Column(
            "tag",
            sa.Enum(
                "ACME_ACCOUNT",
                "CERTIFICATE",
                "COLLECTION",
                "DOMAIN",
                "ENDPOINT",
                "KEY_CHAIN",
                "ROLE",
                "STORAGE_GATEWAY",
                "USER_CREDENTIAL",
                "WILDCARD_DOMAIN",
                name="configurationobjecttag",
            ),
            nullable=False,
        ),
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("serial", sa.INTEGER(), nullable=False),
        sa.Column("configuration", sa.TEXT(), nullable=False),
        sa.Column("deleted", sa.BOOLEAN(), server_default="0", nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute(
        """
        INSERT INTO configuration_object
        (id, created_at, updated_at, discriminator, tag,
         serial, configuration, deleted)
        SELECT id, created_at, updated_at, discriminator, tag,
         serial, configuration,
         case when deleted is null then 0 else 1 end
         from configuration_object_old
        """
    )
    op.drop_table("configuration_object_old")
